/* 
  Lattice2d.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  FEB-1997

*/

#import <random.h>

#import "VNneighborhood.h"
#import "MooreNeighborhood.h"
#import "FHPneighborhood.h"
#import "Coord2d.h"
#import "Lattice2d.h"

@implementation Lattice2d

-(id) setSizeX: (unsigned) x Y: (unsigned) y {
  if ((x % 2) != 0)
    [InternalError raiseEvent:
      "Lattice2d xsize *must* be even.\n"];
  // This is a restriction of the specific way support
  // for the toroidal triangular tiling has been
  // implemented...

  return [super setSizeX: x Y: y];
}

-(Lattice2d *) createEnd {
  Lattice2d * finalSelf;

  finalSelf = [super createEnd];
  finalSelf->tilingCode = SqrTiling;
  finalSelf->neighborhood = 
    [VNneighborhood create: [self getZone]];
  finalSelf->neighborhoodCode = VN;
  finalSelf->rasterScaleFactor = 2; // Default - may not be sensible...

  return finalSelf;
}

+(id) create: (id) aZone withSizeX: (unsigned) sizeX Y: (unsigned) sizeY {
  Lattice2d * lattice2d;

  lattice2d = [self createBegin: aZone];
  [lattice2d setSizeX: sizeX Y: sizeY];
  lattice2d = [lattice2d createEnd];

  return lattice2d;
}

-(void) drop {
  if (neighborhood != nil)
    [neighborhood drop];
  [super drop];
}



/* The following methods provide for processing of
  messages in the <Neighborhood> protocol.  Most of
  this could also be done with a -forward method -
  perhaps more simply and/or efficiently.  But doing it
  this way ensures that there is complete flexibility
  in the method interfaces (any arguments, any return
  type); and also, incidentally, saves the compiler
  from getting upset that the protocol doesn't seem to
  be fully implemented.  Also, there seems to be a
  (benign) declaration conflict for the method
  -performv:, used in forwarding, between Object.h and
  tclObjc.h...  */

-(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
  [neighborhood setUniformUnsigned: inUniformUnsigned];
}

-(id) getUniformUnsigned {
  return uniformUnsigned;
}

-(BOOL) isValidNeighborCode: (NeighborCode) inNeighborCode {
  return [neighborhood isValidNeighborCode: inNeighborCode];
}

-(char *) getNeighborCodeName: (NeighborCode) inNeighborCode {
  return [neighborhood getNeighborCodeName: inNeighborCode];
}

-(OrientationCode) getOrientationCode: (NeighborCode) inNeighborCode {
  return [neighborhood getOrientationCode: inNeighborCode];
}

-(NeighborCode) getInitialNeighborCode {
  return [neighborhood getInitialNeighborCode];
}

-(NeighborCode) getNextNeighborCode: (NeighborCode) inNeighborCode {
  return [neighborhood getNextNeighborCode: inNeighborCode];
}

-(NeighborCode) getPreviousNeighborCode: (NeighborCode) inNeighborCode {
  return [neighborhood getPreviousNeighborCode: inNeighborCode];
}

-(NeighborCode) getOppositeNeighborCode: (NeighborCode) inNeighborCode {
  return [neighborhood getOppositeNeighborCode: inNeighborCode];
}

-(NeighborCode) getRandomNeighborCode {
  return [neighborhood getRandomNeighborCode];
}

-(NeighborCode) getRandomNeighborCodeExcluding: (NeighborCode) inNeighborCode {
  return [neighborhood getRandomNeighborCodeExcluding: inNeighborCode];
}





-(void) setRasterScaleFactor: (unsigned) newScaleFactor {
  rasterScaleFactor = newScaleFactor;
}

-(unsigned) getRasterScaleFactor {
  return rasterScaleFactor;
}

-(TilingCode) getTilingCode; {
  return tilingCode;
}

-(void) setTilingCode: (TilingCode) newTilingCode {
  if (newTilingCode != tilingCode) {
    if (newTilingCode == SqrTiling) {
      tilingCode = newTilingCode;
      neighborhoodCode = VN;
      [neighborhood drop];
      neighborhood = [VNneighborhood create: [self getZone]];
      [neighborhood setUniformUnsigned: uniformUnsigned];
    }
    else if (newTilingCode == TriTiling) {
      tilingCode = newTilingCode;
      neighborhoodCode = FHP;
      [neighborhood drop];
      neighborhood = [FHPneighborhood create: [self getZone]];
      [neighborhood setUniformUnsigned: uniformUnsigned];
    }
    else
      [InternalError raiseEvent:
	"Invalid TilingCode %d.\n", newTilingCode];
  }
}

-(NeighborhoodCode) getNeighborhoodCode {
  return neighborhoodCode;
}

-(void) setNeighborhoodCode: (NeighborhoodCode) newNeighborhoodCode {
  if (newNeighborhoodCode != neighborhoodCode) {
    switch (newNeighborhoodCode) {
    case VN:    
      neighborhoodCode = newNeighborhoodCode;
      [neighborhood drop];
      neighborhood = [VNneighborhood create: [self getZone]];
      [neighborhood setUniformUnsigned: uniformUnsigned];
      tilingCode = SqrTiling;
      break;
    case Moore: 
      neighborhoodCode = newNeighborhoodCode;
      [neighborhood drop];
      neighborhood = [MooreNeighborhood create: [self getZone]];
      [neighborhood setUniformUnsigned: uniformUnsigned];
      tilingCode = SqrTiling;
      break;
    case FHP:   
      neighborhoodCode = newNeighborhoodCode;
      [neighborhood drop];
      neighborhood = [FHPneighborhood create: [self getZone]];
      [neighborhood setUniformUnsigned: uniformUnsigned];
      tilingCode = TriTiling;
      break;
    default:
      [InternalError raiseEvent:
	"Invalid NeighborhoodCode %d.\n", newNeighborhoodCode];
    }
  }
}


-(id) createOriginCoord: (id) aZone {
  Coord2d *coord;

  coord = [Coord2d create: aZone withLattice: self];
  [coord setX: 0];  // Just to be explicit...
  [coord setY: 0];

  return (coord);
}

-(id) createMaxCoord: (id) aZone {
  Coord2d *coord;

  coord = [Coord2d create: aZone withLattice: self];
  [coord setX: (xsize - 1)];
  [coord setY: (ysize - 1)];

  return (coord);
}

-(id) createRandomCoord: (id) aZone {
  Coord2d *coord;

  coord = [Coord2d create: aZone withLattice: self];
  [coord setX: [uniformUnsigned getUnsignedWithMin: 0 withMax: (xsize - 1)]];
  [coord setY: [uniformUnsigned getUnsignedWithMin: 0 withMax: (ysize - 1)]];

  return (coord);
}

@end
