/* 
  LatticeAgent.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  FEB-1997

*/

#import "Lattice.h"
#import "LatticeAgent.h"


@implementation LatticeAgent


-(BOOL) isNeighboringAgentOfClass: (id) someClass {
  BOOL found;
  BOOL done;
  NeighborCode neighborCode, initialNeighborCode;
  LatticeAgent *neighborAgent;
  id lattice;

  done = NO;
  found = NO;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode]; 
  neighborCode = initialNeighborCode;
  while (!done) {
    neighborAgent = [self getAgentAtNeighborCode: neighborCode];
    found = [neighborAgent isMemberOf: someClass];
    neighborCode = 
      [lattice getNextNeighborCode: neighborCode];
    done = (found || (neighborCode == initialNeighborCode));
  }

  return (found);
}

-(id) copyCoord: (id) aZone {
  id newCoord;

  newCoord = nil;
  if (coord != nil)
    newCoord = [coord copy: aZone];

  return (newCoord);
}

-(BOOL) attemptUnwarpToCoord: (id) newCoord {
  id currentOccupant;
  BOOL ok;
   ok = NO;

  if (coord != nil)
    [InternalError raiseEvent: 
      "Attempt to reembed latticeAgent %d;\n"
      "but already embedded.\n",
      self];
  else {
    currentOccupant = [newCoord getObject];
    if (currentOccupant != nil)
      [InternalError raiseEvent: 
	"Attempt to embed latticeAgent %d;\n"
	"but already occupied by agent %d.\n",
	self, currentOccupant];
    else {
      coord = [newCoord copy: [self getZone]];
      [coord putObject: self];
      ok = YES;
    }
  } 
  return ok;
}

-warp 
{

 
  if (coord != nil) {
    [coord putObject: nil];
    [coord drop];
    coord = nil;
  }

  return self;
}


-moveToCoord: (id) newCoord {
  [self warp];
  if (![self attemptUnwarpToCoord: newCoord])
    [InternalError raiseEvent:
      "unwarp failed on attempted move of latticeAgent %d.\n",
      self];

  return self;
}

-moveToNeighborCode: (NeighborCode) neighborCode {
  id newCoord;

  newCoord = [self copyCoord: [self getZone]];
  if (newCoord == nil)
    [InternalError raiseEvent: 
      "Attempt to moveToNeighborCode: on latticeAgent %d,\n"
      "while not embedded.\n",
      self];
  else {
    [newCoord translateByNeighborCode: neighborCode];
    [self moveToCoord: newCoord];
    [newCoord drop];;
  }

  return self;
}


-(BOOL)checkConstraints{ 
//fr LatticeAgent gibt es keine Constraints
return YES;
                       }


//************************************************************************************


-(BOOL) attemptSwapSelfFromOldCoord: (id) oldCoord 
  withAgent: (id) agent
  atNewCoord: (id) newCoord {
  /* Private method; no sanity checks. */
  BOOL primaryOK, secondaryOK, ok;
  
   
    
  [self warp];   
  [agent warp]; 
   
   
//unwarp ist also das zurueckplatzieren des Agenten in der Welt.  
	primaryOK = [self attemptUnwarpToCoord: newCoord];
  	secondaryOK = [agent attemptUnwarpToCoord: oldCoord];
  
    
  
  ok = (primaryOK && secondaryOK);
  
  //hier knnen noch weitere Bedingungen an die Bewegung getestet werden. Ist im Moment nur fr Links und Link2 aktuell (Rasmussen Tests!!)
  
  if(ok){ 
  	 ok = ([self checkConstraints] && [agent checkConstraints]);
  	
	}  
  
  
  
  if (!ok) {
    // Backout!
//printf("\n ---> Backout <----\n");

    [self warp];
    [agent warp];     

     if (![self attemptUnwarpToCoord: oldCoord]) 
      [InternalError raiseEvent:
        "Can't unwarp primary latticeAgent %d back to original coordinate.\n",
        self ];
    if (![agent attemptUnwarpToCoord: newCoord]) 
      [InternalError raiseEvent:
        "Can't unwarp secondary latticeAgent %d back "
		     "to original coordinate).\n",
        agent];
  }

  return ok;
}


-(BOOL) attemptSwapToCoord: (id) newCoord {
  id agent;
  id oldCoord;
  BOOL ok;
  
  ok = NO;
  if (coord == nil)
    [InternalError raiseEvent: 
      "attemptSwapToCoord: on latticeAgent %d,\n"
      "while not embedded.\n",
      self];
  else {
    agent = [newCoord getObject];
    if (agent == nil) {
      [self moveToCoord: newCoord];
      ok = YES;
    }
    else {
      oldCoord = [self copyCoord: [self getZone]];
      ok = ([self attemptSwapSelfFromOldCoord: oldCoord
	withAgent: agent
	atNewCoord: newCoord]);
      [oldCoord drop];
    }
  }

  return ok;
}

-(BOOL) attemptSwapToNeighborCode: (NeighborCode) neighborCode {
  id newCoord;
  BOOL ok;

  ok = NO;

  if (coord == nil)
    [InternalError raiseEvent: 
      "attemptSwapToNeighborCode: on latticeAgent %d,\n"
      "while not embedded.\n",
      self];
  else {
    newCoord = [coord copy: [self getZone]];
    [newCoord translateByNeighborCode: neighborCode];

    ok = ([self attemptSwapToCoord: newCoord]);

    [newCoord drop];
  }

  return (ok);
}

-(BOOL) attemptSwapWithAgent: (LatticeAgent *) agent {
  id newCoord;
  BOOL ok;
  ok = NO;
//Hier finden wir mehr oder weniger nur die Koordinaten des neuen Agenten heraus. Geben dann an attemptswaptocoord ab.
  if (agent == nil)
    [InternalError raiseEvent:
     "attemptSwapWithAgent on latticeAgent %d with nil agent.\n", self];
  else {
    newCoord = [agent copyCoord: [self getZone]];
    if (newCoord == nil)
      [InternalError raiseEvent:
	"attemptSwapWithAgent on latticeAgent %d with latticeAgent%d;\n"
	"but latter not embedded.\n",
	self, agent];
    else {
      ok = [self attemptSwapToCoord: newCoord];
      [newCoord drop];
    }
  }

  return ok;
}


-(id) getAgentAtNeighborCode: (NeighborCode) neighborCode {
  id newCoord;
  id agentAtNeighborCode;

  if (coord == nil) {
    [InternalError raiseEvent: 
      "getAgentAtNeighborCode: on latticeAgent %d,\n"
      "while not embedded.\n",
      self];
    agentAtNeighborCode = nil; // Defensive.
  }
  else {
    newCoord = [coord copy: [self getZone]];
    [newCoord translateByNeighborCode: neighborCode];
    agentAtNeighborCode = [newCoord getObject]; 
    [newCoord drop];
  }
  
// if(agentAtNeighborCode==nil)[InternalError raiseEvent: 
//     "getAgentAtNeighborCode: on latticeAgent %d,\n"
//     "...there is a problem!\n",
//     self];

  return (agentAtNeighborCode);
} 

-(void) drop {
  if (coord != nil) [self warp];
  [super drop];
}

-saveTo: (id <OutFile>) file {
  [super saveTo: file];

  if (coord == nil) {
    [file putInt: 0];
    [file putString: " "];
  }
  else {
    [file putInt: 1];
    [file putString: " "];
    [coord saveTo: file];
  }

  return self;
}


-loadFrom: (id <InFile>) file toLattice: (id) inLattice {
  int embeddedFlag;
  id newCoord;

  [super loadFrom: file];

  [file getInt: &embeddedFlag];
  if (embeddedFlag) {
    newCoord = [inLattice createOriginCoord: [self getZone]];
    [newCoord loadFrom: file];
    
    if (![self attemptUnwarpToCoord: newCoord])
    [InternalError raiseEvent:
      "unwarp failed on attempted load of latticeAgent %d.\n",
      self];
  }

  return self;
}

-(void) xprintStatus {
  printf("LatticeAgent: xprintStatus (self: %p)\n", self);
  printf("  [super xprintStatus]:\n");
  [super xprintStatus];
  if (coord == nil)
    printf("  (coord == nil)\n");
  else {
    printf("  [coord xprintStatus]:\n");
    [coord xprintStatus];
  }
}

@end
