/* 
  Link.h

  Barry McMullin <mcmullin@eeng.dcu.ie>
  27-SEP-1996


*/



#import "Particle.h"
#import "Bond.h"

#import "Bond2.h"
@class Substrate, Bond;

@interface Link : Particle {
  id bondList;
  id bondedLinkList;
  BOOL absorbedSubstrate;
  int linknumber;
  int substratenumber;
  BOOL forceEmission;
  BOOL promoting;
  BOOL disintegrating;
  BOOL splicable;
  unsigned bondedNeighborCount;
}

+createBegin: (id) aZone; // Overridden.
-(void) drop;             // Overridden.

+(void) setupProbeMap: (id) aZone;

// Overridden class methods (see doc in Particle.m)

+(void) setClassMobilityFactor: (double) inMobilityFactor;
+(double) getClassMobilityFactor;

/* This Bernoulli distribution is used in implementing
   the various random reactions.  It must be a
   "variable" or non-"frozen" flavour of Bernoulli,
   created externally to the bonds; in this way a
   single bernoulli object can be shared around
   multiple usages in a single app.  */
+(void) setBernoulli: (id) inBernoulli; 
+(id) getBernoulli;
+(void) setUniformDouble: (id) inUniformDouble; 
+(id) getUniformDouble;
+(void) setUniformUnsigned: (id) inUniformUnsigned; 
+(id) getUniformUnsigned;

// Additional Class methods.
+(void) setLink0MobilityFactor: (double) inMobilityFactor;
+(double) getLink0MobilityFactor;
+(void) setLink1MobilityFactor: (double) inMobilityFactor;
+(double) getLink1MobilityFactor;
+(void) setLink2MobilityFactor: (double) inMobilityFactor;
+(double) getLink2MobilityFactor;
+(void) setLinkPromotionProbability: (double) inPromotionProbability;
+(double) getLinkPromotionProbability;

+(void) setChainInhibitBondFlag: (BOOL) value;
+(BOOL) getChainInhibitBondFlag;
+(void) setCatInhibitBondFlag: (BOOL) value;
+(BOOL) getCatInhibitBondFlag;
+(void) setCatInhibitDisintFlag: (BOOL) value;
+(BOOL) getCatInhibitDisintFlag;
+(void) setAcuteBondingFlag: (BOOL) value;
+(BOOL) getAcuteBondingFlag;
+(void) setLink0SmartMotionFlag: (BOOL) value;
+(BOOL) getLink0SmartMotionFlag;
+(void) setLink1SmartMotionFlag: (BOOL) value;
+(BOOL) getLink1SmartMotionFlag;

+(void) setDisintegrationProbability: (double) inDisintegrationProbability;
+(double) getDisintegrationProbability;

+(void) setNoBondingProbability: (double) probability;
+(double) getNoBondingProbability;
+(void) setChainInitiateProbability: (double) probability;
+(double) getChainInitiateProbability;
+(void) setChainExtendProbability: (double) probability;
+(double) getChainExtendProbability;
+(void) setChainSpliceProbability: (double) probability;
+(double) getChainSpliceProbability;

+(void) setAbsorptionProbability: (double) inAbsorptionProbability;
+(double) getAbsorptionProbability;

+(void) setEmissionProbability: (double) inEmissionProbability;
+(double) getEmissionProbability;

+(void) setActiveRepairProbability: (double) inActiveRepairProbability;
+(double) getActiveRepairProbability;

+(void) setDisplaceProbability: (double) inDisplaceProbability;
+(double) getDisplaceProbability;



// Overridden instance method.
-step;
-(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode;
-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode;
-warp;
-(BOOL) attemptUnwarpToCoord: (id) inCoord;
-(Bond *) createBondToNeighborCode: (NeighborCode) neighborCode ;
-(BOOL) allowBondAtNeighborCode: (NeighborCode) neighborCode;

// Additional instance methods.
-(double) getMobilityFactor;
-(int) getBondCount;
-addBond: (id) bond;
-removeBond: (id) bond;
-(id) getBondAtNeighborCode: (NeighborCode) neighborCode;
 -(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode;
-(void) recordIfSplicable;



-(int) countLinkInNeighborhood;
-(int) countSubstrateInNeighborhood ;
-(BOOL) onlySubstrateNeighbors ;



-(Bond2 *) createBond2ToNeighborCode: (NeighborCode) neighborCode;
-doDisintegration;

/* The following method is intended only for 
   interactive use via a probe. */
-attemptBondAtNeighborCode;


-doBonding;

-(BOOL) isDisintegrating;
-(BOOL) isPromoting;
-(BOOL) hasAbsorbedSubstrate;

-saveTo: (id <OutFile>) file;   // Overridden...
-loadFrom: (id <InFile>) file toLattice: (id) inLattice;  // Overridden...

@end
