/* 
  Link.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  30-SEP-1996

*/

#import <simtools.h>
#import <collections.h>  

#import "SCLGlobals.h"
#import "SCLModelSwarm.h"
#import "LatticeAgent.h"
#import "PickNeighborCode.h"
#import "WorldManager.h"
#import "Bernoulli.h"
#import "Lattice.h"
#import "ParticleManager.h"
#import "Hole.h"
#import "Bond.h"
#import "Bond2.h"

#import "Link.h"
#import "Link2.h"
#import "Substrate.h"
#import "Catalyst.h"

@implementation Link


static id bernoulli;
static id uniformDouble;
static id uniformUnsigned;
static double link0MobilityFactor;
static double link1MobilityFactor;
static double link2MobilityFactor;
static double linkPromotionProbability;
/* Pseudo Class variables. */

+createBegin: (id) aZone {
  Link *link;

  link = [super createBegin: aZone];
  link->bondList = [List create: aZone];
  link->bondedLinkList = [List create: aZone];
 
  return link;
}

-(void) drop {
  [self warp];
  [bondList drop];
  [bondedLinkList drop];
  [super drop];
}

+(void) setBernoulli: (id) inBernoulli {
  bernoulli = inBernoulli;
}

+(id) getBernoulli {
  return(bernoulli);
}

+(void) setUniformDouble: (id) inUniformDouble {
  uniformDouble = inUniformDouble;
}

+(id) getUniformDouble {
  return(uniformDouble);
}

+(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
}

+(id) getUniformUnsigned {
  return(uniformUnsigned);
}

+(void) setupProbeMap: (id) aZone {
  id <ProbeMap> probeMap;

  probeMap = [probeLibrary getProbeMapFor: [self class]];

  [probeMap dropProbeForMessage: "getBondCount"];
  [probeMap dropProbeForMessage: "allowBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "addBond:"];
  [probeMap dropProbeForMessage: "getBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "createBondToNeighborCode:"];
  [probeMap dropProbeForMessage: "removeBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "removeBond:"];
  [probeMap dropProbeForMessage: "doBonding"];
  [probeMap dropProbeForMessage: "doAbsorption"];
  [probeMap dropProbeForMessage: "attemptEmission"];
  [probeMap dropProbeForMessage: "doEmission"];
  [probeMap dropProbeForMessage: "attemptDisintegration"];
  [probeMap dropProbeForMessage: "doDisintegration"];
  [probeMap dropProbeForMessage: "attemptPromotion"];
  [probeMap dropProbeForMessage: "doPromotion"];
  [probeMap dropProbeForMessage: "warp"];
  [probeMap dropProbeForMessage: "doMotion"];
  [probeMap dropProbeForMessage: "canMove"];
  [probeMap dropProbeForMessage: "step"];
  [probeMap dropProbeForMessage: "drawSelfOn:"];
  [probeMap dropProbeForMessage: "saveTo:"];
  [probeMap dropProbeForMessage: "loadFrom:"];
  [probeMap dropProbeForVariable: "absorbedSubstrate"];
  [probeMap dropProbeForVariable: "disintegrating"];
}


+(void) setClassMobilityFactor: (double) inMobilityFactor {
  link0MobilityFactor = inMobilityFactor;
  link1MobilityFactor = inMobilityFactor;
  link2MobilityFactor = inMobilityFactor;
}

+(double) getClassMobilityFactor {
  return(link0MobilityFactor);
}

+(void) setLink0MobilityFactor: (double) inMobilityFactor {
  link0MobilityFactor = inMobilityFactor;
}

+(double) getLink0MobilityFactor {
  return(link0MobilityFactor);
}

+(void) setLink1MobilityFactor: (double) inMobilityFactor {
  link1MobilityFactor = inMobilityFactor;
}

+(double) getLink1MobilityFactor {
  return(link1MobilityFactor);
}

+(void) setLink2MobilityFactor: (double) inMobilityFactor {
  link2MobilityFactor = inMobilityFactor;
}

+(double) getLink2MobilityFactor {
  return(link2MobilityFactor);
}

+(void) setLinkPromotionProbability: (double) inPromotionProbability {
  linkPromotionProbability = inPromotionProbability;
}

+(double) getLinkPromotionProbability {
  return(linkPromotionProbability);
}


static BOOL chainInhibitBondFlag;

+(void) setChainInhibitBondFlag: (BOOL) value {
  chainInhibitBondFlag = value;
}

+(BOOL) getChainInhibitBondFlag {
  return(chainInhibitBondFlag);
}

static BOOL catInhibitBondFlag;

+(void) setCatInhibitBondFlag: (BOOL) value {
  catInhibitBondFlag = value;
}

+(BOOL) getCatInhibitBondFlag {
  return(catInhibitBondFlag);
}

static BOOL acuteBondingFlag;

+(void) setAcuteBondingFlag: (BOOL) value {
  acuteBondingFlag = value;
}

+(BOOL) getAcuteBondingFlag {
  return(acuteBondingFlag);
}

static BOOL catInhibitDisintFlag;

+(void) setCatInhibitDisintFlag: (BOOL) value {
  catInhibitDisintFlag = value;
}

+(BOOL) getCatInhibitDisintFlag {
  return(catInhibitDisintFlag);
}

static BOOL link0SmartMotionFlag;

+(void) setLink0SmartMotionFlag: (BOOL) value {
  link0SmartMotionFlag = value;
}

+(BOOL) getLink0SmartMotionFlag {
  return(link0SmartMotionFlag);
}

static BOOL link1SmartMotionFlag;

+(void) setLink1SmartMotionFlag: (BOOL) value {
  link1SmartMotionFlag = value;
}

+(BOOL) getLink1SmartMotionFlag {
  return(link1SmartMotionFlag);
}

static double disintegrationProbability;

+(void) setDisintegrationProbability: (double) inDisintegrationProbability {
  disintegrationProbability = inDisintegrationProbability;
}

+(double) getDisintegrationProbability {
  return(disintegrationProbability);
}

static double noBondingProbability;

+(void) setNoBondingProbability: (double) probability {
  noBondingProbability = probability;
}

+(double) getNoBondingProbability {
  return(noBondingProbability);
}

static double chainInitiateProbability;

+(void) setChainInitiateProbability: (double) probability {
  chainInitiateProbability = probability;
}

+(double) getChainInitiateProbability {
  return(chainInitiateProbability);
}

static double chainExtendProbability;

+(void) setChainExtendProbability: (double) probability {
  chainExtendProbability = probability;
}

+(double) getChainExtendProbability {
  return(chainExtendProbability);
}

static double chainSpliceProbability;

+(void) setChainSpliceProbability: (double) probability {
  chainSpliceProbability = probability;
}

+(double) getChainSpliceProbability {
  return(chainSpliceProbability);
}

static double absorptionProbability;

+(void) setAbsorptionProbability: (double) inAbsorptionProbability {
  absorptionProbability = inAbsorptionProbability;
}

+(double) getAbsorptionProbability {
  return(absorptionProbability);
}

static double emissionProbability;

+(void) setEmissionProbability: (double) inEmissionProbability {
  emissionProbability = inEmissionProbability;
}

+(double) getEmissionProbability {
  return(emissionProbability);
}

static double activeRepairProbability;

+(void) setActiveRepairProbability: (double) inActiveRepairProbability {
  activeRepairProbability = inActiveRepairProbability;
}

+(double) getActiveRepairProbability {
  return(activeRepairProbability);
}

static double displaceProbability;

+(void) setDisplaceProbability: (double) inDisplaceProbability {
  displaceProbability = inDisplaceProbability;
}

+(double) getDisplaceProbability {
  return(displaceProbability);
}

-(BOOL) isDisintegrating {
  return (disintegrating);
}

-(BOOL) isPromoting {
  return (promoting);
}

-(BOOL) hasAbsorbedSubstrate {
  return (absorbedSubstrate);
}

-(int) getBondCount {
  return [bondList getCount];
}


-(BOOL) neighboringChain {
  /* 
    This is loosely inspired by the PROTOBIO/exp27.for 
    code. Returns YES if there is at least 1 doubly bonded
    link in the neighborhood of self...
  */
  BOOL chainFound;
  BOOL done;
  NeighborCode neighborCode, initialNeighborCode;
  LatticeAgent *neighborAgent;
  Link *neighborLink;
  id lattice;

  done = NO;
  chainFound = NO;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode]; 
  neighborCode = initialNeighborCode;
  while (!done) {
    neighborAgent = [self getAgentAtNeighborCode: neighborCode];
    if ([neighborAgent isMemberOf: [Link class]]) {
      neighborLink = (Link *) neighborAgent;
      if ([neighborLink getBondCount] == 2) 
        chainFound = YES;
    }
    neighborCode = 
      [lattice getNextNeighborCode: neighborCode];
    done = (chainFound || (neighborCode == initialNeighborCode));
  }

  return (chainFound);
}

-(BOOL) mayBond {
  BOOL inhibit;

  inhibit = ([bondList getCount] >= 2);

  if (!inhibit && ([bondList getCount] == 0)) 
    inhibit = 
      ([Link getChainInhibitBondFlag] && [self neighboringChain]) ||
      ([Link getCatInhibitBondFlag] && 
        [self isNeighboringAgentOfClass: [Catalyst class]]);

  return (!inhibit);
}

-(BOOL) allowBondAtNeighborCode: (NeighborCode) neighborCode {
  /* You can't bond where there is an existing bond, or
     at the next or previous neighborCodes (latter 
     qualified by acuteBondingFlag). */
 //  Particle* neighborParticle;
//   Link* link = nil;
//   Link2* link2 = nil;
  BOOL inhibitedByAcuteBonding;
  BOOL allow;
  inhibitedByAcuteBonding = NO;
  if (!acuteBondingFlag)
    inhibitedByAcuteBonding = 
      ([self getBondAtNeighborCode: 
	[[coord getLatticeObject] 
	  getNextNeighborCode: neighborCode]] != nil) ||
      ([self getBondAtNeighborCode: 
	[[coord getLatticeObject] 
	  getPreviousNeighborCode: neighborCode]] != nil);

  allow =  ((!inhibitedByAcuteBonding ) &&
    ([self getBondAtNeighborCode: neighborCode] == nil));

return allow;

}

 

-addBond: (id) bond {
  /* This method *should* only be invoked if it is
     already known that it is "allowed" to add the
     requested bond; there are *no* checks here!!!
      
     Grundsaetzlich wollen wir hier als Eingabe nur Bond und Bond2 */
  [bondList addLast: bond];

  return self;
}

-removeBondAtNeighborCode: (NeighborCode) neighborCode {
  id index;
  id bond;
   Bond* bond1=nil;
   Bond2* bond2=nil;
  BOOL found;
  NeighborCode hilfsneighborCode=Center;
 
  found = NO;
  index = [bondList begin: [self getZone]];
  bond = [index next];
  while ((bond != nil) && (!found)) {
  
  if([bond isMemberOf: [Bond2 class]]){
  bond2 = (Bond2*)bond;
  hilfsneighborCode = [bond getNeighborCodeForLink: self];}
  
  if([bond isMemberOf: [Bond class]]){
  bond1 = (Bond*) bond;
  hilfsneighborCode = [bond1 getNeighborCodeForLink: self];}
  				
    if (hilfsneighborCode == neighborCode) {
      found = YES;
      [index remove];
    }
    else 
      bond = [index next];
  }
  [index drop];

  if (!found) 
    [InternalError raiseEvent:
      "Attempt to remove non-existent bond on link %p at neighborCode %d.\n",
      self, neighborCode];

  return self;
}

-(id) getBondAtNeighborCode: (NeighborCode) neighborCode {
  //hier fragen wir ob es in eine bestimmte Richtung einen Bond gibt. Gibt den Bond zurueck wenn ja.
  
  id index;
  id bond; 
  BOOL found;
  Bond* bond1=nil;
  Bond2* bond2=nil;

  found = NO;
  index = [bondList begin: [self getZone]];
  bond = [index next];
   
  
    
  while ((bond != nil) && (!found)) {
   if(bond!=nil) //Eigentlich berflssig, macht aber die Methode mehr symmetrisch
   if([bond isMemberOf: [Bond class]]){
   bond1 = (Bond*) bond;
    if ([bond1 getNeighborCodeForLink: self] == neighborCode)
      found = YES;
     					}
      
      
      if(bond!=nil)
      if([bond isMemberOf: [Bond2 class]]){
   bond2 = (Bond2*) bond;
    if ([bond2 getNeighborCodeForLink: self] == neighborCode)
      found = YES;
     					}	
          bond = [index next];
          }
  [index drop];
  

  return bond;
}

 


-(Bond *) createBondToNeighborCode: (NeighborCode) neighborCode {
  Bond *bond;

//Das hier wird verwendet um  Bond mit Link zu machen!

  /*
  if (![self mayBond])
    [WarningMessage raiseEvent: 
      "Creating \"inhibited\" bond at neighborCode %d of link %d.\n",
      neighborCode, self];
      */
  
  
  if (![self allowBondAtNeighborCode: neighborCode]) 
    [WarningMessage raiseEvent: 
      "Creating \"disallowed\" bond at neighborCode %d of link %d.\n",
      neighborCode, self];

  bond = [bondMgr createAgent];
  [bond setLink0: self];
  [bond setNeighbor0: neighborCode];
  [bond setOrientationCode: 
    [[coord getLatticeObject] getOrientationCode: neighborCode]];
  [self addBond: bond];

  return bond;
}




-(Bond2 *) createBond2ToNeighborCode: (NeighborCode) neighborCode {
  Bond2 *bond;

  //Diese Methode wird aufgerufen um Bond2 mit einem Link2 zu bilden!
  
  if (![self allowBondAtNeighborCode: neighborCode]) 
    [WarningMessage raiseEvent: 
      "Creating \"disallowed\" bond at neighborCode %d of link %d.\n",
      neighborCode, self];

 if(![[self getAgentAtNeighborCode: neighborCode] isMemberOf: [Link2 class]])
    [WarningMessage raiseEvent: 
      "Neighbor is not Link2 %d of link %d.\n",
      neighborCode, self];

 
 
  bond = [bond2Mgr createAgent];
  [bond setLink0: self];
  [bond setNeighbor0: neighborCode];
  [bond setOrientationCode: 
    [[coord getLatticeObject] getOrientationCode: neighborCode]];
  [self addBond: bond];

  return bond;
}

-attemptBondAtNeighborCode {
  /* This is intended only for user interaction, via a probe... */
  Particle *neighborParticle;
  Link *neighborLink;
  Bond* bond;
  NeighborCode neighborCode, oppositeNeighborCode;
  id pickNeighborCode;
 

   pickNeighborCode = 
    [PickNeighborCode create: [self getZone]
		      withLattice: [coord getLatticeObject]];
  neighborCode = [pickNeighborCode getNeighborCode];
  [pickNeighborCode drop];

  if ([self allowBondAtNeighborCode: neighborCode]) {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Link class]]) {
	neighborLink = (Link *) neighborParticle;
        oppositeNeighborCode = 
	  [[coord getLatticeObject] getOppositeNeighborCode: neighborCode];
        if ([neighborLink mayBond] && 
            [neighborLink allowBondAtNeighborCode: oppositeNeighborCode]) {
	  bond = [neighborLink createBondToNeighborCode: oppositeNeighborCode];
	  [bond setLink1: self];
	  [bond setNeighbor1: neighborCode];
	  [self addBond: bond];
	}
      }
    }
  }



 
  [worldManager update];
  return self;
}



-removeBond: (id) bond {
  id removedBond;

   removedBond = [bondList remove: bond];

  if (removedBond != bond) 
    [InternalError raiseEvent:
      "Attempt to remove non-existent bond %p on link %p.\n",
      bond, self];

  return self;
}



typedef enum {
  noBonding, chainInitiate, chainExtend, chainSplice
} BondAction;

typedef struct {
  id particle;
  NeighborCode toCode;
  NeighborCode fromCode;
 } NeighborRecord;

-(void) scanForBondableNeighbors: (id) bondableNeighbors
  // bondableNeighbors must be an empty List.
  // NB! Wir suchen hier nur nach Links, nicht nach Link2
{
  NeighborCode toNeighborCode, fromNeighborCode, initialNeighborCode;
  NeighborRecord *neighborRecord;
  id neighborParticle;
  id lattice;


   lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  toNeighborCode = initialNeighborCode;
  do {
    if ([self allowBondAtNeighborCode: toNeighborCode]) {
      neighborParticle = [self getAgentAtNeighborCode: toNeighborCode];

      if (neighborParticle == nil) 
	[InternalError raiseEvent:
	   "nil encountered at neighborCode %d of particle %d.\n",
	   toNeighborCode, self];
      else {
	if ([neighborParticle isMemberOf: [Link class]]) {
	  if ([neighborParticle mayBond]) {
	    fromNeighborCode = 
	      [lattice getOppositeNeighborCode: toNeighborCode];
	    if ([neighborParticle allowBondAtNeighborCode: 
		fromNeighborCode]) {
	      neighborRecord = [[self getZone] alloc: sizeof(NeighborRecord)];
	      neighborRecord->particle = neighborParticle;
	      neighborRecord->toCode = toNeighborCode;
	      neighborRecord->fromCode = fromNeighborCode;
	      [bondableNeighbors addLast: (id) neighborRecord];
	    }
	  }
	}
      }
    }
    toNeighborCode = [lattice getNextNeighborCode: toNeighborCode];
  } while (toNeighborCode != initialNeighborCode);
}






-(BOOL) onlySubstrateNeighbors {


  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice; int i = 0;
  int SubstrateCount = 0; 
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do { i++;
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
      
    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Substrate class]]) 
	  SubstrateCount++;
      } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
  if(i>SubstrateCount)
  return NO;
  else return YES;
}













-(void) dropNeighborList: (id) aList {
  id index;
  NeighborRecord *neighborRecord;

  index = [aList begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    [index remove];
    [[self getZone] free: (id) neighborRecord];
  }
  [index drop];
  [aList drop];
}    

-(id) pickRandomListMember: (id) aList {
  unsigned count, randOffset;
  id member;

   count = [aList getCount];
  member = nil;

  if (count > 0) {
    if (count == 1)
      randOffset = 0;
    else
      randOffset = [uniformUnsigned getUnsignedWithMin: 0
		    withMax: (count - 1)];

    member = [aList atOffset: randOffset];
  }

  return(member);
}


-(Bond*) formBondToNeighbor: (NeighborRecord *) neighborRecord {
  Bond *bond = nil;  
  Link* link;
  Link2* link2;
  

   if([neighborRecord->particle isMemberOf: [Link class]]){
  link = (Link*) neighborRecord->particle;
   
  bond = [link  createBondToNeighborCode: neighborRecord->fromCode];}

//Dieser Fall sollte eigentlich nie eintreffen, habe aber den Eindruck, dass der Compiler es so will!!!

  if([neighborRecord->particle isMemberOf: [Link2 class]]){
 [InternalError raiseEvent:  	 "Link attempts to form Bond with Link2" ];
  link2 = (Link2*) neighborRecord->particle;
  bond = (Bond*)[link2  createBondToNeighborCode: neighborRecord->fromCode];}



  [bond setLink1: self];
  [bond setNeighbor1: neighborRecord->toCode];
  [self addBond: bond];

  return(bond);
}

-(void) attemptChainInitiate {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;
    if ([self mayBond] && ([bondList getCount] == 0)) {
    bondableNeighbors = [List create: [self getZone]];
    [self scanForBondableNeighbors: bondableNeighbors];

    // Limit list to *free* Links...
    index = [bondableNeighbors begin: [self getZone]];
    while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
      if ([neighborRecord->particle getBondCount] != 0) {
	[index remove];
	[[self getZone] free: neighborRecord];
      }
    }
    [index drop];

    // Limit list to  Link2... Rasmussen Condition
    index = [bondableNeighbors begin: [self getZone]];
    while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
      if (![neighborRecord->particle isMemberOf: [Link2 class]]) {
	[index remove];
	[[self getZone] free: neighborRecord];
      }
    }
    [index drop];



//In neighborRecord stehen nun alle Nachbarn ohne Bond drinnen! Nur Link2
    neighborRecord = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    if (neighborRecord != NULL)
      [self formBondToNeighbor: neighborRecord];

    [self dropNeighborList: bondableNeighbors];
  }
}


-(void) attemptChainExtend {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;
  unsigned myCount;

    if ([self mayBond]) {
    bondableNeighbors = [List create: [self getZone]];
    [self scanForBondableNeighbors: bondableNeighbors];

    // Limit list to potential extensions...
    myCount = [bondList getCount];
    index = [bondableNeighbors begin: [self getZone]];
    while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
      if ((([neighborRecord->particle getBondCount] + myCount) != 1) || 
/*Rasmussen Bedingung:*/  ![neighborRecord->particle isMemberOf: [Link class]]) {
	[index remove];
	[[self getZone] free: neighborRecord];
      }
    }
    [index drop];

    neighborRecord = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    if (neighborRecord != NULL)
      {[self formBondToNeighbor: neighborRecord]; }

    [self dropNeighborList: bondableNeighbors];
  }
}

-(void) attemptFreeChainSplice {
  NeighborRecord *neighborRecord1, *neighborRecord2;
  id bondableNeighbors;
  id index;
  id bond1;

   

  bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord1 = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord1->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord1];
    }
  }
  [index drop];

  if ([bondableNeighbors getCount] >= 2) {
    neighborRecord1 = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    [bondableNeighbors remove: (id) neighborRecord1];
    neighborRecord2 =
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    [bondableNeighbors addLast: (id) neighborRecord1];

    bond1 = [self formBondToNeighbor: neighborRecord1];

    if ( [self mayBond] &&
      [self allowBondAtNeighborCode: neighborRecord2->toCode] &&
      [neighborRecord2->particle mayBond] &&
      [neighborRecord2->particle 
	allowBondAtNeighborCode: neighborRecord2->fromCode]) {
	  [self formBondToNeighbor: neighborRecord2];
	  [worldManager logSplice];
    }
    else {
      // Second bond incompatible with first...
      [bond1 clearLinks];
      [bondMgr dropAgent: bond1];  
    }
  }

  [self dropNeighborList: bondableNeighbors];
}

-(void) attemptBondedChainSplice {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;

   bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord];
    }
  }
  [index drop];

  neighborRecord = 
    (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
  if (neighborRecord != NULL) {
    [self formBondToNeighbor: neighborRecord];
    [worldManager logSplice];
  }

  [self dropNeighborList: bondableNeighbors];
}

-(void) attemptChainSplice {
  if ([self mayBond]) {
    if ([bondList getCount] == 0) [self attemptFreeChainSplice];
    else [self attemptBondedChainSplice];
  }
}

-(BondAction) chooseBondAction {
  BondAction action;
  double sample; 
  sample = [uniformDouble getDoubleWithMin: 0.0 withMax: 1.0];
  if (sample < noBondingProbability)
    action = noBonding;
  else {
    sample -= noBondingProbability;
    if (sample < chainInitiateProbability)
      action = chainInitiate;
    else {
      sample -= chainInitiateProbability;
      if (sample < chainExtendProbability)
	action = chainExtend;
	//We just do not need chain Splice. In case, it is no-Bonding
      else  action = noBonding;
           
	//action = chainSplice;
    }
  }

  return(action);
}

-doBonding {
  BondAction bondAction;

  if (!promoting) {
    bondAction = [self chooseBondAction];
    switch (bondAction) {
      case noBonding: break;
      case chainInitiate: [self attemptChainInitiate]; break;
      case chainExtend:   [self attemptChainExtend];   break;
      case chainSplice:   [self attemptChainSplice];   break;
      default:
    }
  }

  return self;
}


-dropBonds {
  id bond;
Bond* bond1= nil;
Bond2* bond2 = nil;
   while ([bondList getCount] > 0) {
    bond = [bondList getFirst];
    //Hier mu wieder diese etwas anstrengende Fallunterscheidung durchgefuehrt werden.
    
    if([bond isMemberOf: [Bond class]]){
    bond1 = (Bond*)bond;
    [bond1 clearLinks];
    [bondMgr dropAgent: bond1];}
    
    if([bond isMemberOf: [Bond2 class]]){
    bond2 = (Bond2*)bond;
    [bond2 clearLinks];
    [bond2Mgr dropAgent: bond2];}
    
    if((![bond isMemberOf: [Bond class]]) && (![bond isMemberOf: [Bond2 class]]) && bond != nil)
    [InternalError raiseEvent: "Inconsistency with call: -drop Bonds in Link.m object: %d.\n" , self];
  }

  return self;
}

-doAbsorption {
  NeighborCode neighborCode;
  Particle *neighborParticle;
  id neighborCoord;
  Hole *hole;
 
  if (!absorbedSubstrate) {
    if ([bernoulli getTossWithProb: absorptionProbability]) {
      neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
      neighborParticle = [self getAgentAtNeighborCode: neighborCode];
      if (neighborParticle == nil)
	[InternalError raiseEvent:
	   "nil encountered at neighborCode %d of particle %d.\n",
	   neighborCode, self];
      else {
	if ([neighborParticle isMemberOf: [Substrate class]]) {
	  neighborCoord = [neighborParticle copyCoord: [self getZone]];
	  [substrateMgr dropAgent: (Substrate *) neighborParticle];
	  hole = [holeMgr createAgent];
	  if (![hole attemptUnwarpToCoord: neighborCoord])
	    [InternalError raiseEvent:
	      "Failed unwarp of hole %d.\n", hole];
	  [neighborCoord drop];
          absorbedSubstrate = YES;
	}
      }
    }
  }

  return self;
}

-attemptEmission {
  NeighborCode neighborCode;
  Particle *neighborParticle;
  id neighborCoord;
  Substrate *substrate;

  if (!absorbedSubstrate)
    [InternalError raiseEvent:
       "attemptEmission when no absorbed substrate?\n"];
  else {
    neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
    if (neighborParticle == nil)
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Hole class]]) {
	neighborCoord = [neighborParticle copyCoord: [self getZone]];
	[holeMgr dropAgent: (Hole *) neighborParticle];
	substrate = [substrateMgr createAgent];
	if (![substrate attemptUnwarpToCoord: neighborCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate %d.\n", substrate];
	[neighborCoord drop];
	absorbedSubstrate = NO;
      }
    }
  }

  return self;
}

-doEmission {
  if (absorbedSubstrate) 
    if ([bernoulli getTossWithProb: emissionProbability]) 
      [self attemptEmission];

  return self;
}


-attemptDisintegration {
  NeighborCode neighborCode;
  Particle *neighborParticle;
  id myCoord, neighborCoord;
  Substrate *substrate0, *substrate1;

  if (!absorbedSubstrate) {
    neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
    if (neighborParticle == nil)
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Hole class]]) {
	/* We can only *actually* disintegrate if we have a hole to drop
	   the extra substrate into... */
	neighborCoord = [neighborParticle copyCoord: [self getZone]];
	[holeMgr dropAgent: (Hole *) neighborParticle];
	substrate0 = [substrateMgr createAgent];
	if (![substrate0 attemptUnwarpToCoord: neighborCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate0 %d.\n", substrate0];
	[neighborCoord drop];

	myCoord = [self copyCoord: [self getZone]];
	[linkMgr dropAgent: self];
	substrate1 = [substrateMgr createAgent];
	if (![substrate1 attemptUnwarpToCoord: myCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate0 %d.\n", substrate0];
	[myCoord drop];
      }
    }
  }

  return self;
}

-doDisintegration {
  if (catInhibitDisintFlag && 
    [self isNeighboringAgentOfClass: [Catalyst class]]) 
      disintegrating = NO;
  else {
    if ((!disintegrating) && (!promoting)) 
      if ([bernoulli getTossWithProb: disintegrationProbability]) {
        disintegrating = YES;
        forceEmission = absorbedSubstrate;
      }
  }

  return self;
}




-warp {
  id hilfsbond;
 Bond* bond = nil;
  Bond2* bond2 = nil;
   
 

 if (coord != nil) {
    [bondedLinkList removeAll]; // Defensive...
    while ([bondList getCount] > 0) {
      // Note that we will be removing all members from
      // bondList (indirectly, via the [bond
      // clearLinks] message); so it is *correct* to
      // repeatedly call [bondList getFirst] to get the
      // *different* members.  It would be positively
      // wrong to use an index to traverse bondList
      // here, because the index validity would be
      // compromised through the indirect modification
      // of the list by [bond clearLinks]..
         
      
       hilfsbond = [bondList getFirst];  
       
        if([hilfsbond isMemberOf: [Bond class]]){
      bond = (Bond *) hilfsbond;
      [bondedLinkList addLast: [bond getLinkOppositeTo: self]];
      [bond clearLinks];
       [bondMgr dropAgent: bond];                }
 
       if([hilfsbond isMemberOf: [Bond2 class]]) {
       bond2 = (Bond2 *) hilfsbond;
      [bondedLinkList addLast: [bond2 getLinkOppositeTo: self]];
      [bond2 clearLinks];
       [bond2Mgr dropAgent: bond2];         }
      
    }
    [super warp];
  } 
  return self;
}


-(BOOL) attemptUnwarpToCoord: (id) inCoord {
  int doneCount;
  

  NeighborCode initialNeighborCode, neighborCode, oppositeNeighborCode;
  id neighborAgent = nil;
  Link* neighborLink=nil;  
  Link2* neighborLink2=nil;  
  
  Bond* bond;
  Bond2* bond2;
  BOOL done, result;
  id lattice;
  
  //initialisiere, damit sich der Compiler nicht beklagt!
  bond = nil; 
  bond2= nil;

   
  if (![super attemptUnwarpToCoord: inCoord])
    [InternalError raiseEvent:
      "failed super unwarp of %d.\n", self];

  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  doneCount = 0;
  
   done = (doneCount >= [bondedLinkList getCount]);
    
  while (!done) {  
  
  neighborAgent = [self getAgentAtNeighborCode: neighborCode];
     
    
     //Wenn neighborAgent in der Liste der gebundenen Teiclhen ist, dann...
    if ([bondedLinkList contains: neighborAgent]) {//if3
      
      doneCount++;
      if ([self allowBondAtNeighborCode: neighborCode]) {//if2
 
	
	//hier muessen wir eine kleine Fallunterscheidung machen, weil bonding mit Link 2 nach einer anderen creatBondToNeighborcode Funktion verlangt. Samt Fehlerueberpruefung.



    if([neighborAgent isMemberOf: [Link2 class]]) {
	neighborLink2 = (Link2 *) neighborAgent; 
 
	oppositeNeighborCode = 
	  [lattice getOppositeNeighborCode: neighborCode];
	if ([neighborLink2 allowBondAtNeighborCode: oppositeNeighborCode]) {//if1
	
   	   bond2 =  [neighborLink2 createBondToNeighborCode: oppositeNeighborCode];
	    	 
	  [bond2 setLink21: self];  
	  [bond2 setNeighbor1: neighborCode];
	  [self addBond: bond2];                          
 									 }//Ende if1
	 
							} //Fall Link2 Ende!
 
//Jetzt kommt der Fall Link 1
	
    if([neighborAgent isMemberOf:   [Link class]]){
      neighborLink = (Link *) neighborAgent;
      
 	oppositeNeighborCode = 
	  [lattice getOppositeNeighborCode: neighborCode];
	if ([neighborLink allowBondAtNeighborCode: oppositeNeighborCode]) {//if1
	
	    
	  bond =   [neighborLink createBondToNeighborCode: oppositeNeighborCode];	 
	   
	  [bond setLink1: self];  
	  [bond setNeighbor1: neighborCode];
	  [self addBond: bond];                          
 
	}//Ende if1
	
      }//Ende Fall Link
      }//Ende if2
    } //Ende if3
    
    
    if (doneCount >= [bondedLinkList getCount]) 
      done = YES;
    else {
      neighborCode = 
	[lattice getNextNeighborCode: neighborCode];
	    done = (neighborCode == initialNeighborCode);
    }
  }//Ende while

  result = NO;

//Hier wird kontrolliert ob der Link die Bewegung zult nmlich indirekt dadurch, da kontrolliert wird 
// ob in die vorige Richtung noch ein Bond geschlossen werden kann!!

   
  if ([bondList getCount] == [bondedLinkList getCount]) {
    result = YES;
    [bondedLinkList removeAll];  // Defensive...
  }
  else {
    result = NO;
    /* NB: we can't use [self warp] here because we would
       lose the bondedLinkList info... */
    [self dropBonds];
    [super warp];
  }

 
  return result;
}


-(BOOL)checkConstraints{ 
 
 
 //Rasmussen Regel: Konrolliere ob in der neuen Konfiguration auch zumindest genausoviele Links sind!
   //and dieser Stelle gibt es noch die Moeglichkeit da eine der Nachbarzellen noch nicht unwarped ist: Das macht aber nichts, weil ddiese sicher keinen Link enthalten wird: Der Gruind dafr ist, da Links mict sich selber nicht Platzs tauschen.
   
   //Linkophil
  // if(linknumber >  [self countLinkInNeighborhood]) return  NO;
   
   //hydrophob
   if(substratenumber <  [self countSubstrateInNeighborhood]) return  NO;


else return YES;
                       }


-(int) countLinkInNeighborhood {
//Diese Methode ist fast identisch mit countSubstrateNeighbors in Link2!!!

  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;
  int LinkCount; 
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  LinkCount = 0;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
if(neighborParticle != nil)
       if ([neighborParticle isMemberOf: [Link class]]) 
	  LinkCount++;
     neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
  
  return LinkCount;

}



-(int) countSubstrateInNeighborhood {
//Diese Methode ist fast identisch mit countSubstrateNeighbors in Link2!!!

  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;
  int SubstrateCount; 
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  SubstrateCount = 0;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
if(neighborParticle != nil)
       if ([neighborParticle isMemberOf: [Substrate class]]) 
	  SubstrateCount++;
     neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
  
  return SubstrateCount;

}



-(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode{

//Link Zwei laesst sich nur von Substrat bewegen wenn es ausschliesslich Substrate in seiner Nachbarschaft gibt. Sonst geht Bewegung nicht..


//Particle *neighborParticle;
BOOL canMove;


canMove =  [super canBeMovedToNeighborCode: neighborCode];



//Diese Regel brauch wir nicht mehr! Wird jetzt fuer Rasmusseregel eingetauscht. 

//neighborParticle = [self getAgentAtNeighborCode: neighborCode];
 
//if ([neighborParticle isMemberOf: [Substrate class]] && ![self onlySubstrateNeighbors])
 //canMove = NO;
 
 
 if([self getBondCount] > 0) canMove = NO;

return canMove;


}




/*    Bewegung  Tests ob Bedingung erfuellt domotion*/
-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode {
  BOOL canMove;
  Particle *neighborParticle;



 //Registriere als erstes wie viele Substrate und Links es in der Nachbarschaft gibt!
 
 linknumber = [self countLinkInNeighborhood];  

 substratenumber  = [self countSubstrateInNeighborhood];  

   
  canMove = [super canMoveToNeighborCode:  neighborCode];
 

  if (canMove) {
    if ([bondList getCount] > 0) {
      if ([self getBondAtNeighborCode: neighborCode] != nil)
	canMove = NO;
 neighborParticle = [self getAgentAtNeighborCode: neighborCode];
	canMove = 
	  ([neighborParticle isMemberOf: [Hole class]] || 
	   [neighborParticle isMemberOf: [Link2 class]]|| 
	   [neighborParticle isMemberOf: [Catalyst class]]|| 
	   [neighborParticle isMemberOf: [Substrate class]]); 
      
 
 }  
 else{   
 neighborParticle = [self getAgentAtNeighborCode: neighborCode];	
 if(![self onlySubstrateNeighbors]){
	canMove = 
	  ([neighborParticle isMemberOf: [Hole class]] ||
	    [neighborParticle isMemberOf: [Link2 class]] ||
	    [neighborParticle isMemberOf: [Catalyst class]]||
	    [neighborParticle isMemberOf: [Substrate class]] );
                                    }   
      }//Ende else 
    
   } 
 
 
 
  return canMove;


  }
 







// -(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode {
//   BOOL canMove;
//   
//   Particle *neighborParticle; 
//   canMove = [super canMoveToNeighborCode:  neighborCode];
// 
//  if (canMove) {
//     if ([bondList getCount] > 0) {
//       if ([self getBondAtNeighborCode: neighborCode] != nil)
// 	canMove = NO;
//  neighborParticle = [self getAgentAtNeighborCode: neighborCode];
// 	canMove = 
// 	  ([neighborParticle isMemberOf: [Hole class]] || 
// 	   [neighborParticle isMemberOf: [Link2 class]]|| 
// 	   [neighborParticle isMemberOf: [Catalyst class]]); 
//       
//  
//  }  
//     
//  //Soferne es andere als Substrate in der Nachbarschaft gibt, laesst sich Link2  nicht auf eine Bewegung mit Substrat ein, sondern nur mit Hole und Link. Gibt es in seiner Nachbarschaft allerdings nur Substrate, dannn  sind die Fuer eine Bewegung gut genug.     
//  else{   
//  neighborParticle = [self getAgentAtNeighborCode: neighborCode];	
//  if(![self onlySubstrateNeighbors]){
// 	canMove = 
// 	  ([neighborParticle isMemberOf: [Hole class]] ||
// 	    [neighborParticle isMemberOf: [Link2 class]] ||
// 	    [neighborParticle isMemberOf: [Catalyst class]] );
//                                     }   
//       }//Ende else 
//     
//    } 
// 
// 
//  
// 
//   
//    return canMove;
// }
// 
// 
// 
// -(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode{
// //Link  laesst sich nur von Substrat bewegen ausser es gibt keines in seiner Nachbarschaft.
// 
// 
// Particle *neighborParticle;
// BOOL canMove;
// 
// 
// canMove =  [super canBeMovedToNeighborCode: neighborCode];
// 
// 
// neighborParticle = [self getAgentAtNeighborCode: neighborCode];
// 
//  //hier wird codiert, da sich Link bevorzugt mit Substrat bewegt
// if (![neighborParticle isMemberOf: [Substrate class]] && ([self onlySubstrateNeighbors] != 0))
//  { canMove = NO;}
//   
// 
// //Wenn ein Link Bonds hat kann er nicht bewegt werden!
// 
//  //if([self getBondCount] > 0) canMove = NO;
//  
//  
// return canMove;
// }



-(void) recordIfSplicable {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;

  bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord];
    }
  }
  [index drop];

  if ([bondList getCount] == 0)
   splicable = ([bondableNeighbors getCount] >= 2);
  else
   splicable = ([bondableNeighbors getCount] >= 1);

  [self dropNeighborList: bondableNeighbors];
}


-(void) countBondedNeighbors {
  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;

  bondedNeighborCount = 0U;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Link class]]) {
	switch ([neighborParticle getBondCount]) {
	case 1:
	case 2:
	  bondedNeighborCount++;
	  break;
	default:
	}
      }
    } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
}

-(void) prepareForMotion {
  switch ([bondList getCount]) {

  case 0:
    if (link0SmartMotionFlag) {
      [self recordIfSplicable];
      [self countBondedNeighbors];
    }
    break;

  case 1:
    if (link1SmartMotionFlag) 
      [self recordIfSplicable];
    break;

  case 2:
  default:
  }
}


-(double) getMobilityFactor {
  double mobilityFactor;
  int bondCount;
  unsigned oldBondedNeighborCount;
  BOOL oldSplicableFlag;

  bondCount = [bondList getCount];

  switch (bondCount) {

  case 0: 
    mobilityFactor = link0MobilityFactor; 
    if (link0SmartMotionFlag) {
      oldSplicableFlag = splicable;
      [self recordIfSplicable];

      oldBondedNeighborCount = bondedNeighborCount;
      [self countBondedNeighbors];

      if (oldSplicableFlag && !splicable) 
	mobilityFactor = 0.0; // Veto move!
      else
	if ((!splicable) && (!disintegrating) &&
	   (oldBondedNeighborCount > bondedNeighborCount))
	  mobilityFactor = 0.0; // Veto move!
    }
    break;

  case 1: 
    mobilityFactor = link1MobilityFactor; 
    if (link1SmartMotionFlag) {
       oldSplicableFlag = splicable;
       [self recordIfSplicable];

       if (oldSplicableFlag && !splicable) 
 	mobilityFactor = 0.0; // Veto move!
    }
    break;

  case 2: 
    mobilityFactor = link2MobilityFactor; 
    break;

  default:
      [InternalError raiseEvent:
       "Out-of-range bondCount (%d) on Link (%u).\n", 
       [bondList getCount], self];
      mobilityFactor = 0.0; // Defensive!
  }

  return mobilityFactor;
}


-(void) scanForRepairableNeighbors: (id) repairableList {
  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;

  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Link class]]) {
	if ([neighborParticle isDisintegrating] &&
	    ([neighborParticle getBondCount] > 0))
	  [repairableList addLast: neighborParticle];
      }
    } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
}

-(void) doRepairMotion {
  id repairableLinkList;
  id repairableLink;

  if ((!hasMoved) && (!disintegrating) && ([bondList getCount] == 0)) {
    if ([bernoulli getTossWithProb: activeRepairProbability]) {
      repairableLinkList = [List create: [self getZone]];
      [self scanForRepairableNeighbors: repairableLinkList];
      repairableLink = [self pickRandomListMember: repairableLinkList];
      [repairableLinkList drop];
      if (repairableLink != nil) {
	[repairableLink dropBonds];
	hasMoved = [self attemptSwapWithAgent: repairableLink];
	if (!hasMoved)
	  [InternalError raiseEvent:
	    "Attempted motion for repair failed?\n"];
	else {
	  [repairableLink setHasMoved];
	  [worldManager logRepairMotion];
	}
      }
    }
  }
}


-(BOOL) displaceInDirection: (NeighborCode) displaceDirection {

// 
// 
// id bond;
//   NeighborCode oldBondFromDirection, oldBondToDirection;
//   id oldBondPartner;
//   BOOL displaced;
// 
//   displaced = NO;
// 
//   if (disintegrating)
//     [InternalError raiseEvent:
//       "Attempted displacement, but disintegrating?\n"];
// 
//   if ([bondList getCount] != 2)
//     [InternalError raiseEvent:
//       "Attempted displacement, but (bondCount != 2)?\n"];
// 
//   if ([self canMoveToNeighborCode: displaceDirection]) {
//     if ([bernoulli getTossWithProb: 0.5]) 
//       bond = [bondList atOffset: 0];
//     else
//       bond = [bondList atOffset: 1];
// 
//     oldBondPartner = [bond getLinkOppositeTo: self];
//     oldBondFromDirection = [bond getNeighborCodeForLink: oldBondPartner];
//     oldBondToDirection = [bond getNeighborCodeForLink: self];
//     [bond clearLinks];
//     [bondMgr dropAgent: bond];
// 
//     hasMoved = [self attemptSwapToNeighborCode: displaceDirection];
//     if (hasMoved) {
//       [[self getAgentAtNeighborCode: displaceDirection] setHasMoved];
//       displaced = YES;
//     }
//     else {
//       // Reform the bond that was broken...
//       bond = [oldBondPartner createBondToNeighborCode: oldBondFromDirection];
//       [bond setLink1: self];
//       [bond setNeighbor1: oldBondToDirection];
//       [self addBond: bond];
//     }
//   }
//   return(displaced);

return NO;


}

-(void) scanForDisplaceableNeighbors: (id) displaceableNeighbors
  // displaceableNeighbors must be an empty List.
{
  NeighborCode toNeighborCode, initialNeighborCode;
  NeighborRecord *neighborRecord;
  id neighborParticle;
  id lattice;

  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  toNeighborCode = initialNeighborCode;
  do {
    if ([self canMoveToNeighborCode: toNeighborCode]) {
      neighborParticle = [self getAgentAtNeighborCode: toNeighborCode];

      if (neighborParticle == nil) 
	[InternalError raiseEvent:
	   "nil encountered at neighborCode %d of particle %d.\n",
	   toNeighborCode, self];
      else {
	if ([neighborParticle isMemberOf: [Link class]]) {
	  if (([neighborParticle getBondCount] == 2) && 
              (![neighborParticle isDisintegrating])) {
	    neighborRecord = [[self getZone] alloc: sizeof(NeighborRecord)];
	    neighborRecord->particle = neighborParticle;
	    neighborRecord->toCode = toNeighborCode;
	    neighborRecord->fromCode = 
              [lattice getOppositeNeighborCode: toNeighborCode];
	    [displaceableNeighbors addLast: (id) neighborRecord];
	  }
	}
      }
    }
    toNeighborCode = [lattice getNextNeighborCode: toNeighborCode];
  } while (toNeighborCode != initialNeighborCode);
}


-(void) doDisplaceMotion {
  id displaceableLinkList;
  NeighborRecord *neighborRecord;
  NeighborCode displaceDirection;
  id displaceableLink;

  if ((!hasMoved) && (!disintegrating) && 
      ([bondList getCount] == 0) && (displaceProbability != 0.0)) {
    if ([bernoulli getTossWithProb: displaceProbability]) {
      displaceableLinkList = [List create: [self getZone]];
      [self scanForDisplaceableNeighbors: displaceableLinkList];
      neighborRecord = 
	(NeighborRecord *) [self pickRandomListMember: displaceableLinkList];
      if (neighborRecord != NULL) {
	displaceableLink = neighborRecord->particle;
	displaceDirection = neighborRecord->toCode;
	if ([displaceableLink displaceInDirection: displaceDirection]) {
	  hasMoved = [self attemptSwapToNeighborCode: displaceDirection];
	  if (!hasMoved)
	    [InternalError raiseEvent:
	      "Attempted motion after displacement failed?\n"];
	  else {
	    [[self getAgentAtNeighborCode: displaceDirection] setHasMoved];  
	    [worldManager logDisplaceMotion];
	  }
	}
      }
      [self dropNeighborList: displaceableLinkList];
    }
  }
}

-attemptPromotion {
  id myCoord, cat;

  if (!absorbedSubstrate)
    {
      myCoord = [self copyCoord: [self getZone]];
      [linkMgr dropAgent: self];
      cat = [catalystMgr createAgent];
      if (![cat attemptUnwarpToCoord: myCoord])
	[InternalError raiseEvent:
	  "Failed unwarp of cat %d.\n", link];
      [myCoord drop];
    }
  
  return self;
}

-doPromotion {
  if ((!promoting) &&
      (!disintegrating) && 
      ([bondList getCount] == 0) && 
      (linkPromotionProbability != 0.0))
    if ([self isNeighboringAgentOfClass: [Catalyst class]]) 
    // Promotion must be catalysed; the idea is to
    // prevent spontaneous promotion out in the
    // environment, which would just be a distraction and
    // possibly interfere with the cell proper...
      if ([bernoulli getTossWithProb: linkPromotionProbability]) {
        promoting = YES;
        forceEmission = absorbedSubstrate;
      }

  return self;
}

 
-step {
  //[self doRepairMotion]; 
  //[self doDisplaceMotion];
  [self doMotion];
  [self doBonding];
  [self doDisintegration];
  [self doPromotion];

  if (forceEmission) {
    [self attemptEmission];
    forceEmission = absorbedSubstrate;
  }

  if (disintegrating)
    [self attemptDisintegration];
  else 
    if (promoting)
      [self attemptPromotion];
    else {
      [self doEmission];
      [self doAbsorption];
    }

  return self;
}


-saveTo: (id <OutFile>) file {
  /* Note that we do *not* save bond information here; that info
     is already implicit in the bond agent list; of course, it
     would be more conservative to save the redundancies and then
     crosscheck on loading - but, for the time being at least,
     I judge that to be an unwarrananted complication... */

  [super saveTo: file];

  [file putInt: absorbedSubstrate];
  [file putString: " "];
  [file putInt: disintegrating];  
  [file putString: " "];

  return self;
}

-loadFrom: (id <InFile>) file toLattice: (id) inLattice {
  int intFlag;

  [super loadFrom: file toLattice: inLattice];

  [file getInt: &intFlag];
  if ((intFlag != YES) && (intFlag != NO))
    [InternalError raiseEvent:
      "Loaded BOOL value (%d) not YES or NO.", intFlag];
  absorbedSubstrate = (BOOL) intFlag; 


  [file getInt: &intFlag];
  if ((intFlag != YES) && (intFlag != NO))
    [InternalError raiseEvent:
      "Loaded BOOL value (%d) not YES or NO.", intFlag];
   disintegrating = (BOOL) intFlag;  

  return self;
}


@end
