/* 
  Link2.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  30-SEP-1996

*/

#import <simtools.h>
#import <collections.h>  

#import "SCLGlobals.h"
#import "SCLModelSwarm.h"

#import "WorldManager.h"
#import "Bernoulli.h"
#import "Lattice.h"
#import "ParticleManager.h"
#import "Hole.h"
#import "Link.h"
#import "Substrate.h"
#import "Bond.h"

#import "Bond2.h"

#import "Catalyst.h"
#import "PickNeighborCode.h"
#import "Link2.h"

@implementation Link2


static id bernoulli;
static id uniformDouble;
static id uniformUnsigned;
static double link0MobilityFactor;
static double link1MobilityFactor;
static double link2MobilityFactor;
static double linkPromotionProbability;
/* Pseudo Class variables. */

+createBegin: (id) aZone {
  Link2 *link2;

  link2 = [super createBegin: aZone];
  link2->bondList = [List create: aZone];
  link2->bondedLinkList = [List create: aZone];

  return link2;
}

-(void) drop {
  [self warp];
  [bondList drop];
  [bondedLinkList drop];
  [super drop];
}

+(void) setBernoulli: (id) inBernoulli {
  bernoulli = inBernoulli;
}

+(id) getBernoulli {
  return(bernoulli);
}

+(void) setUniformDouble: (id) inUniformDouble {
  uniformDouble = inUniformDouble;
}

+(id) getUniformDouble {
  return(uniformDouble);
}

+(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
}

+(id) getUniformUnsigned {
  return(uniformUnsigned);
}

+(void) setupProbeMap: (id) aZone {
  id <ProbeMap> probeMap;

  probeMap = [probeLibrary getProbeMapFor: [self class]];

  [probeMap dropProbeForMessage: "getBondCount"];
  [probeMap dropProbeForMessage: "allowBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "addBond:"];
  [probeMap dropProbeForMessage: "getBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "createBondToNeighborCode:"];
  [probeMap dropProbeForMessage: "removeBondAtNeighborCode:"];
  [probeMap dropProbeForMessage: "removeBond:"];
  [probeMap dropProbeForMessage: "doBonding"];
  [probeMap dropProbeForMessage: "doAbsorption"];
  [probeMap dropProbeForMessage: "attemptEmission"];
  [probeMap dropProbeForMessage: "doEmission"];
  [probeMap dropProbeForMessage: "attemptDisintegration"];
  [probeMap dropProbeForMessage: "doDisintegration"];
  [probeMap dropProbeForMessage: "attemptPromotion"];
  [probeMap dropProbeForMessage: "doPromotion"];
  [probeMap dropProbeForMessage: "warp"];
  [probeMap dropProbeForMessage: "doMotion"];
  [probeMap dropProbeForMessage: "canMove"];
  [probeMap dropProbeForMessage: "step"];
  [probeMap dropProbeForMessage: "drawSelfOn:"];
  [probeMap dropProbeForMessage: "saveTo:"];
  [probeMap dropProbeForMessage: "loadFrom:"];
  [probeMap dropProbeForVariable: "absorbedSubstrate"];
  [probeMap dropProbeForVariable: "disintegrating"];
}


+(void) setClassMobilityFactor: (double) inMobilityFactor {
  link0MobilityFactor = inMobilityFactor;
  link1MobilityFactor = inMobilityFactor;
  link2MobilityFactor = inMobilityFactor;
}

+(double) getClassMobilityFactor {
  return(link0MobilityFactor);
}

+(void) setLink0MobilityFactor: (double) inMobilityFactor {
  link0MobilityFactor = inMobilityFactor; 
}

+(double) getLink0MobilityFactor {
  return(link0MobilityFactor);
}

+(void) setLink1MobilityFactor: (double) inMobilityFactor {
  link1MobilityFactor = inMobilityFactor;
}

+(double) getLink1MobilityFactor {
  return(link1MobilityFactor);
}

+(void) setLink2MobilityFactor: (double) inMobilityFactor {
  link2MobilityFactor = inMobilityFactor;
}

+(double) getLink2MobilityFactor {
  return(link2MobilityFactor);
}

+(void) setLinkPromotionProbability: (double) inPromotionProbability {
  linkPromotionProbability = inPromotionProbability;
}

+(double) getLinkPromotionProbability {
  return(linkPromotionProbability);
}


static BOOL chainInhibitBondFlag;

+(void) setChainInhibitBondFlag: (BOOL) value {
  chainInhibitBondFlag = value;
}

+(BOOL) getChainInhibitBondFlag {
  return(chainInhibitBondFlag);
}

static BOOL catInhibitBondFlag;

+(void) setCatInhibitBondFlag: (BOOL) value {
  catInhibitBondFlag = value;
}

+(BOOL) getCatInhibitBondFlag {
  return(catInhibitBondFlag);
}

static BOOL acuteBondingFlag;

+(void) setAcuteBondingFlag: (BOOL) value {
  acuteBondingFlag = value;
}

+(BOOL) getAcuteBondingFlag {
  return(acuteBondingFlag);
}

static BOOL catInhibitDisintFlag;

+(void) setCatInhibitDisintFlag: (BOOL) value {
  catInhibitDisintFlag = value;
}

+(BOOL) getCatInhibitDisintFlag {
  return(catInhibitDisintFlag);
}

static BOOL link0SmartMotionFlag;

+(void) setLink0SmartMotionFlag: (BOOL) value {
  link0SmartMotionFlag = value;
}

+(BOOL) getLink0SmartMotionFlag {
  return(link0SmartMotionFlag);
}

static BOOL link1SmartMotionFlag;

+(void) setLink1SmartMotionFlag: (BOOL) value {
  link1SmartMotionFlag = value;
}

+(BOOL) getLink1SmartMotionFlag {
  return(link1SmartMotionFlag);
}

static double disintegrationProbability;

+(void) setDisintegrationProbability: (double) inDisintegrationProbability {
  disintegrationProbability = inDisintegrationProbability;
}

+(double) getDisintegrationProbability {
  return(disintegrationProbability);
}

static double noBondingProbability;

+(void) setNoBondingProbability: (double) probability {
  noBondingProbability = probability;
}

+(double) getNoBondingProbability {
  return(noBondingProbability);
}

static double chainInitiateProbability;

+(void) setChainInitiateProbability: (double) probability {
  chainInitiateProbability = probability;
}

+(double) getChainInitiateProbability {
  return(chainInitiateProbability);
}

static double chainExtendProbability;

+(void) setChainExtendProbability: (double) probability {
  chainExtendProbability = probability;
}

+(double) getChainExtendProbability {
  return(chainExtendProbability);
}

static double chainSpliceProbability;

+(void) setChainSpliceProbability: (double) probability {
  chainSpliceProbability = probability;
}

+(double) getChainSpliceProbability {
  return(chainSpliceProbability);
}

static double absorptionProbability;

+(void) setAbsorptionProbability: (double) inAbsorptionProbability {
  absorptionProbability = inAbsorptionProbability;
}

+(double) getAbsorptionProbability {
  return(absorptionProbability);
}

static double emissionProbability;

+(void) setEmissionProbability: (double) inEmissionProbability {
  emissionProbability = inEmissionProbability;
}

+(double) getEmissionProbability {
  return(emissionProbability);
}

static double activeRepairProbability;

+(void) setActiveRepairProbability: (double) inActiveRepairProbability {
  activeRepairProbability = inActiveRepairProbability;
}

+(double) getActiveRepairProbability {
  return(activeRepairProbability);
}

static double displaceProbability;

+(void) setDisplaceProbability: (double) inDisplaceProbability {
  displaceProbability = inDisplaceProbability;
}

+(double) getDisplaceProbability {
  return(displaceProbability);
}

-(BOOL) isDisintegrating {
  return (disintegrating);
}

-(BOOL) isPromoting {
  return (promoting);
}

-(BOOL) hasAbsorbedSubstrate {
  return (absorbedSubstrate);
}

-(int) getBondCount {
  return [bondList getCount];
}


-(BOOL) neighboringChain {
  /* 
    This is loosely inspired by the PROTOBIO/exp27.for 
    code. Returns YES if there is at least 1 doubly bonded
    link in the neighborhood of self...
  */
  BOOL chainFound;
  BOOL done;
  NeighborCode neighborCode, initialNeighborCode;
  LatticeAgent *neighborAgent;
  Link2 *neighborLink;
  id lattice;

  done = NO;
  chainFound = NO;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode]; 
  neighborCode = initialNeighborCode;
  while (!done) {
    neighborAgent = [self getAgentAtNeighborCode: neighborCode];
    if ([neighborAgent isMemberOf: [Link2 class]]) {
      neighborLink = (Link2 *) neighborAgent;
      if ([neighborLink getBondCount] == 2) 
        chainFound = YES;
    }
    neighborCode = 
      [lattice getNextNeighborCode: neighborCode];
    done = (chainFound || (neighborCode == initialNeighborCode));
  }

  return (chainFound);
}

-(BOOL) mayBond {
  BOOL inhibit;

  inhibit = ([bondList getCount] >= 2);

  if (!inhibit && ([bondList getCount] == 0)) 
    inhibit = 
      ([Link2 getChainInhibitBondFlag] && [self neighboringChain]) ||
      ([Link2 getCatInhibitBondFlag] && 
        [self isNeighboringAgentOfClass: [Catalyst class]]);

  return (!inhibit);
}

-(BOOL) allowBondAtNeighborCode: (NeighborCode) neighborCode {
  /* You can't bond where there is an existing bond, or
     at the next or previous neighborCodes (latter 
     qualified by acuteBondingFlag). */
  BOOL inhibitedByAcuteBonding;

  inhibitedByAcuteBonding = NO;
  if (!acuteBondingFlag)
    inhibitedByAcuteBonding = 
      ([self getBondAtNeighborCode: 
	[[coord getLatticeObject] 
	  getNextNeighborCode: neighborCode]] != nil) ||
      ([self getBondAtNeighborCode: 
	[[coord getLatticeObject] 
	  getPreviousNeighborCode: neighborCode]] != nil);

  return ((!inhibitedByAcuteBonding ) &&
    ([self getBondAtNeighborCode: neighborCode] == nil));
}


-addBond: (Bond2 *) bond {
  /* This method *should* only be invoked if it is
     already known that it is "allowed" to add the
     requested bond; there are *no* checks here!!! */
  [bondList addLast: bond];

  return self;
}

-removeBondAtNeighborCode: (NeighborCode) neighborCode {
  id index;
 Bond2* bond;
  BOOL found;

  found = NO;
  index = [bondList begin: [self getZone]];
  bond = [index next];
  while ((bond != nil) && (!found)) {
    if ([bond getNeighborCodeForLink: self] == neighborCode) {
      found = YES;
      [index remove];
    }
    else 
      bond = [index next];
  }
  [index drop];

  if (!found) 
    [InternalError raiseEvent:
      "Attempt to remove non-existent bond on link2 %p at neighborCode %d.\n",
      self, neighborCode];

  return self;
}

-(Bond2 *) getBondAtNeighborCode: (NeighborCode) neighborCode {
  id index;
  Bond2* bond;
  BOOL found;

  found = NO;
  index = [bondList begin: [self getZone]];
  bond = [index next];
  while ((bond != nil) && (!found)) {
    if ([bond getNeighborCodeForLink: self] == neighborCode)
      found = YES;
    else 
      bond = [index next];
  }
  [index drop];

  return bond;
}

-(Bond2 *) createBondToNeighborCode: (NeighborCode) neighborCode {
 Bond2 *bond ;
 
 bond = nil;

  
 if (![self allowBondAtNeighborCode: neighborCode]) 
    [WarningMessage raiseEvent: 
      "Creating \"disallowed\" bond at neighborCode %d of link %d.\n",
      neighborCode, self];

  bond = [bond2Mgr createAgent];
  [bond setLink20: self];
  [bond setNeighbor0: neighborCode];
  [bond setOrientationCode: 
    [[coord getLatticeObject] getOrientationCode: neighborCode]];
  [self addBond: bond];

  return bond;
}


-attemptBondAtNeighborCode {
  /* This is intended only for user interaction, via a probe... */
//   Particle *neighborParticle;
//   Link2 *neighborLink;
//   Bond2 *bond;
//   NeighborCode neighborCode, oppositeNeighborCode;
//   id pickNeighborCode;
// 
//   pickNeighborCode = 
//     [PickNeighborCode create: [self getZone]
// 		      withLattice: [coord getLatticeObject]];
//   neighborCode = [pickNeighborCode getNeighborCode];
//   [pickNeighborCode drop];
// 
//   if ([self allowBondAtNeighborCode: neighborCode]) {
//     neighborParticle = [self getAgentAtNeighborCode: neighborCode];
//     if (neighborParticle == nil) 
//       [InternalError raiseEvent:
// 	 "nil encountered at neighborCode %d of particle %d.\n",
// 	 neighborCode, self];
//     else {
//       if ([neighborParticle isMemberOf: [Link2 class]]) {
// 	neighborLink = (Link2 *) neighborParticle;
//         oppositeNeighborCode = 
// 	  [[coord getLatticeObject] getOppositeNeighborCode: neighborCode];
//         if ([neighborLink mayBond] && 
//             [neighborLink allowBondAtNeighborCode: oppositeNeighborCode]) {
// 	  bond = [neighborLink createBondToNeighborCode: oppositeNeighborCode];
// 	  [bond setLink1: self];
// 	  [bond setNeighbor1: neighborCode];
// 	  [self addBond: bond];
// 	}
//       }
//     }
//   }
// 
//   [worldManager update];
  return self;
}

-removeBond: (Bond2 *) bond {
  id removedBond;

  removedBond = [bondList remove: bond];

  if (removedBond != bond) 
    [InternalError raiseEvent:
      "Attempt to remove non-existent bond %p on link %p.\n",
      bond, self];

  return self;
}


typedef enum {
  noBonding, chainInitiate, chainExtend, chainSplice
} BondAction;

typedef struct {
  id particle;
  NeighborCode toCode;
  NeighborCode fromCode;
} NeighborRecord;

-(void) scanForBondableNeighbors: (id) bondableNeighbors
  // bondableNeighbors must be an empty List.
{
  NeighborCode toNeighborCode, fromNeighborCode, initialNeighborCode;
  NeighborRecord *neighborRecord;
  id neighborParticle;
  id lattice;

  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  toNeighborCode = initialNeighborCode;
  do {
    if ([self allowBondAtNeighborCode: toNeighborCode]) {
      neighborParticle = [self getAgentAtNeighborCode: toNeighborCode];

      if (neighborParticle == nil) 
	[InternalError raiseEvent:
	   "nil encountered at neighborCode %d of particle %d.\n",
	   toNeighborCode, self];
      else {
	if ([neighborParticle isMemberOf: [Link class]]) {
	  if ([neighborParticle mayBond]) {
	    fromNeighborCode = 
	      [lattice getOppositeNeighborCode: toNeighborCode];
	    if ([neighborParticle allowBondAtNeighborCode: 
		fromNeighborCode]) {
	      neighborRecord = [[self getZone] alloc: sizeof(NeighborRecord)];
	      neighborRecord->particle = neighborParticle;
	      neighborRecord->toCode = toNeighborCode;
	      neighborRecord->fromCode = fromNeighborCode;
	      [bondableNeighbors addLast: (id) neighborRecord];
	    }
	  }
	}
      }
    }
    toNeighborCode = [lattice getNextNeighborCode: toNeighborCode];
  } while (toNeighborCode != initialNeighborCode);
}

-(void) dropNeighborList: (id) aList {
  id index;
  NeighborRecord *neighborRecord;

  index = [aList begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    [index remove];
    [[self getZone] free: (id) neighborRecord];
  }
  [index drop];
  [aList drop];
}    

-(id) pickRandomListMember: (id) aList {
  unsigned count, randOffset;
  id member;

  count = [aList getCount];
  member = nil;

  if (count > 0) {
    if (count == 1)
      randOffset = 0;
    else
      randOffset = [uniformUnsigned getUnsignedWithMin: 0
		    withMax: (count - 1)];

    member = [aList atOffset: randOffset];
  }

  return(member);
}


-(Bond2 *) formBondToNeighbor: (NeighborRecord *) neighborRecord {
Bond2 *bond;
Link* link; 

//Im Moment sind nur Bonds mit Link zugelassen. Das mu modifiziert werden, wenn ein Link2 auch mit einem anderen Link2 sich verbinden kann.

link = neighborRecord->particle ;
  bond = [link  createBond2ToNeighborCode: neighborRecord->fromCode];
 if(![bond isMemberOf: [Bond2 class]]) {printf("\n Verdammt!!!! \n"); abort();}	   
 
  [bond setLink1:  self];
  [bond setNeighbor1: neighborRecord->toCode];
  [self addBond: bond];

  return(bond);
}

-(void) attemptChainInitiate {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;

  if ([self mayBond] && ([bondList getCount] == 0)) {
    bondableNeighbors = [List create: [self getZone]];
    [self scanForBondableNeighbors: bondableNeighbors];

    // Limit list to *free* Links...
    index = [bondableNeighbors begin: [self getZone]];
    while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
      if ([neighborRecord->particle getBondCount] != 0) {
	[index remove];
	[[self getZone] free: neighborRecord];
      }
    }
    [index drop];

    neighborRecord = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    if (neighborRecord != NULL)
      [self formBondToNeighbor: neighborRecord];

    [self dropNeighborList: bondableNeighbors];
  }
}


-(void) attemptChainExtend {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;
  unsigned myCount;

  if ([self mayBond]) {
    bondableNeighbors = [List create: [self getZone]];
    [self scanForBondableNeighbors: bondableNeighbors];

    // Limit list to potential extensions...
    myCount = [bondList getCount];
    index = [bondableNeighbors begin: [self getZone]];
    while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
      if (([neighborRecord->particle getBondCount] + myCount) != 1) {
	[index remove];
	[[self getZone] free: neighborRecord];
      }
    }
    [index drop];

    neighborRecord = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    if (neighborRecord != NULL)
      [self formBondToNeighbor: neighborRecord];

    [self dropNeighborList: bondableNeighbors];
  }
}

-(void) attemptFreeChainSplice {
  NeighborRecord *neighborRecord1, *neighborRecord2;
  id bondableNeighbors;
  id index;
  id bond1;

printf("splice");
  bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord1 = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord1->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord1];
    }
  }
  [index drop];

  if ([bondableNeighbors getCount] >= 2) {
    neighborRecord1 = 
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    [bondableNeighbors remove: (id) neighborRecord1];
    neighborRecord2 =
      (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
    [bondableNeighbors addLast: (id) neighborRecord1];

    bond1 = [self formBondToNeighbor: neighborRecord1];

    if ( [self mayBond] &&
      [self allowBondAtNeighborCode: neighborRecord2->toCode] &&
      [neighborRecord2->particle mayBond] &&
      [neighborRecord2->particle 
	allowBondAtNeighborCode: neighborRecord2->fromCode]) {
	  [self formBondToNeighbor: neighborRecord2];
	  [worldManager logSplice];
    }
    else {
      // Second bond incompatible with first...
      [bond1 clearLinks];
      [bond2Mgr dropAgent: bond1];  
    }
  }

  [self dropNeighborList: bondableNeighbors];
}

-(void) attemptBondedChainSplice {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;


printf("splice");

  bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord];
    }
  }
  [index drop];

  neighborRecord = 
    (NeighborRecord *)[self pickRandomListMember: bondableNeighbors];
  if (neighborRecord != NULL) {
    [self formBondToNeighbor: neighborRecord];
    [worldManager logSplice];
  }

  [self dropNeighborList: bondableNeighbors];
}

-(void) attemptChainSplice {printf("splice");
  if ([self mayBond]) {
    if ([bondList getCount] == 0) [self attemptFreeChainSplice];
    else [self attemptBondedChainSplice];
  }
}

-(BondAction) chooseBondAction {
  BondAction action;
  double sample;
  sample = [uniformDouble getDoubleWithMin: 0.0 withMax: 1.0];

  if (sample < noBondingProbability)
    action = noBonding;
  else {
    sample -= noBondingProbability;
    if (sample < chainInitiateProbability)
      action = chainInitiate;
    else {
      sample -= chainInitiateProbability;
      if (sample < chainExtendProbability)
	action = chainExtend;
      else
	action = chainSplice;
    }
  }

  return(action);
}

-doBonding {
  BondAction bondAction;


  if (!promoting) { 
    bondAction = [self chooseBondAction];
    switch (bondAction) {
      case noBonding: break;
      case chainInitiate: [self attemptChainInitiate]; break;
      case chainExtend:   [self attemptChainExtend];   break;
      case chainSplice:   [self attemptChainSplice];   break;
      default:
    }
  }

  return self;
}


-dropBonds {
  Bond2 *bond;

  while ([bondList getCount] > 0) {
    bond = [bondList getFirst];
    [bond clearLinks];
    [bond2Mgr dropAgent: bond];
  }

  return self;
}

-doAbsorption {// 
//   NeighborCode neighborCode;
//   Particle *neighborParticle;
//   id neighborCoord;
//   Hole *hole;
// 
//   if (!absorbedSubstrate) {
//     if ([bernoulli getTossWithProb: absorptionProbability]) {
//       neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
//       neighborParticle = [self getAgentAtNeighborCode: neighborCode];
//       if (neighborParticle == nil)
// 	[InternalError raiseEvent:
// 	   "nil encountered at neighborCode %d of particle %d.\n",
// 	   neighborCode, self];
//       else {
// 	if ([neighborParticle isMemberOf: [Substrate class]]) {
// 	  neighborCoord = [neighborParticle copyCoord: [self getZone]];
// 	  [substrateMgr dropAgent: (Substrate *) neighborParticle];
// 	  hole = [holeMgr createAgent];
// 	  if (![hole attemptUnwarpToCoord: neighborCoord])
// 	    [InternalError raiseEvent:
// 	      "Failed unwarp of hole %d.\n", hole];
// 	  [neighborCoord drop];
//           absorbedSubstrate = YES;
// 	}
//       }
//     }
//   }

  return self;
}

-attemptEmission {
  NeighborCode neighborCode;
  Particle *neighborParticle; 
  id neighborCoord;
  Substrate *substrate;

  if (!absorbedSubstrate)
    [InternalError raiseEvent:
       "attemptEmission when no absorbed substrate?\n"];
  else {
    neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
    if (neighborParticle == nil)
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Hole class]]) {
	neighborCoord = [neighborParticle copyCoord: [self getZone]];
	[holeMgr dropAgent: (Hole *) neighborParticle];
	substrate = [substrateMgr createAgent];
	if (![substrate attemptUnwarpToCoord: neighborCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate %d.\n", substrate];
	[neighborCoord drop];
	absorbedSubstrate = NO;
      }
    }
  }

  return self;
}

-doEmission {//  
//   if (absorbedSubstrate) 
//     if ([bernoulli getTossWithProb: emissionProbability]) 
//       [self attemptEmission];

  return self;
}


-attemptDisintegration {
  NeighborCode neighborCode;
  Particle *neighborParticle;
  id myCoord, neighborCoord;
  Substrate *substrate0, *substrate1;

  if (!absorbedSubstrate) {
    neighborCode = [[coord getLatticeObject] getRandomNeighborCode];
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
    if (neighborParticle == nil)
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Hole class]]) {
	/* We can only *actually* disintegrate if we have a hole to drop
	   the extra substrate into... */
	neighborCoord = [neighborParticle copyCoord: [self getZone]];
	[holeMgr dropAgent: (Hole *) neighborParticle];
	substrate0 = [substrateMgr createAgent];
	if (![substrate0 attemptUnwarpToCoord: neighborCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate0 %d.\n", substrate0];
	[neighborCoord drop];

	myCoord = [self copyCoord: [self getZone]];
	[link2Mgr dropAgent: self];
	substrate1 = [substrateMgr createAgent];
	if (![substrate1 attemptUnwarpToCoord: myCoord])
	  [InternalError raiseEvent:
	    "Failed unwarp of substrate0 %d.\n", substrate0];
	[myCoord drop];
      }
    }
  }

  return self;
}

-doDisintegration { // 
//   if (catInhibitDisintFlag && 
//     [self isNeighboringAgentOfClass: [Catalyst class]]) 
//       disintegrating = NO;
//   else {
//     if ((!disintegrating) && (!promoting)) 
//       if ([bernoulli getTossWithProb: disintegrationProbability]) {
//         disintegrating = YES;
//         forceEmission = absorbedSubstrate;
//       }
//   }

  return self;
}




-(BOOL)checkConstraints{ 
 
 
 //Rasmussen Regel: Konrolliere ob in der neuen Konfiguration auch zumindest genausoviele Links sind!
   //and dieser Stelle gibt es noch die Moeglichkeit da eine der Nachbarzellen noch nicht unwarped ist:  
   
   //Substratenumber wird in canMovetoNeighborcode registriert!
   
   
   //hydrophil
   if(substratenumber >  [self countSubstrateNeighbors]) return  NO;


else return YES;
                       }




-warp {
  Bond2 *bond;

  if (coord != nil) {
    [bondedLinkList removeAll]; // Defensive...
    while ([bondList getCount] > 0) {
      // Note that we will be removing all members from
      // bondList (indirectly, via the [bond
      // clearLinks] message); so it is *correct* to
      // repeatedly call [bondList getFirst] to get the
      // *different* members.  It would be positively
      // wrong to use an index to traverse bondList
      // here, because the index validity would be
      // compromised through the indirect modification
      // of the list by [bond clearLinks]..
      bond = [bondList getFirst];
      [bondedLinkList addLast: [bond getLink2OppositeTo: self]];
      [bond clearLinks];
      [bond2Mgr dropAgent: bond];
    }
    [super warp];
  }
  return self;
}


-(BOOL) attemptUnwarpToCoord: (id) inCoord {
  int doneCount;
  NeighborCode initialNeighborCode, neighborCode, oppositeNeighborCode;
  id neighborAgent = nil;
  Link* neighborLink = nil;
  Link2* neighborLink2 = nil;
  Bond2 *bond = nil;
  BOOL done, result;
  id lattice;

  if (![super attemptUnwarpToCoord: inCoord])
    [InternalError raiseEvent:
      "failed super unwarp of %d.\n", self];

  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  doneCount = 0;
  done = (doneCount >= [bondedLinkList getCount]);
  while (!done) {
    neighborAgent = [self getAgentAtNeighborCode: neighborCode];
    if ([bondedLinkList contains: neighborAgent]) {//if3
      doneCount++;
      if ([self allowBondAtNeighborCode: neighborCode]) { //if2
       
      
	  
 
 
 //Fall Nachbar ist auch Link2 
 	   if([neighborAgent isMemberOf: [Link2 class]]){
 
	neighborLink2 = (Link2 *) neighborAgent;  

	oppositeNeighborCode = 
	  [lattice getOppositeNeighborCode: neighborCode];
	if ([neighborLink2 allowBondAtNeighborCode: oppositeNeighborCode]) { //if 1
	
	 
	bond = 
	    [neighborLink2 createBondToNeighborCode: oppositeNeighborCode];
							}//Ende if1

							}//Ende falls Link2	
	



//Fall Nachbar ist Link	
	
	    if([neighborAgent isMemberOf: [Link class]]){
	
	   neighborLink = (Link *) neighborAgent;

	oppositeNeighborCode = 
	  [lattice getOppositeNeighborCode: neighborCode];
	if ([neighborLink allowBondAtNeighborCode: oppositeNeighborCode]) { //if 1
	 
	
	  bond = 
	    [neighborLink createBond2ToNeighborCode: oppositeNeighborCode];
	    //Fehlercheck
	    if(bond==nil) abort();
	    						} //Ende if1
 

	  [bond setLink1: self];
	  [bond setNeighbor1: neighborCode];
	  [self addBond: bond];
										} //Falls Link
      
      
      
      }//Ende if2
    }//Ende if3
    if (doneCount >= [bondedLinkList getCount]) 
      done = YES;
    else {
      neighborCode = 
	[lattice getNextNeighborCode: neighborCode];
      done = (neighborCode == initialNeighborCode);
    }  //Ende else
  } //Ende while Schleife

  result = NO;
  if ([bondList getCount] == [bondedLinkList getCount]) {
    result = YES;
    [bondedLinkList removeAll];  // Defensive...
  }
  else {
    result = NO;
    /* NB: we can't use [self warp] here because we would
       lose the bondedLinkList info... */
    [self dropBonds];
    [super warp];
  }

  return result;
}
 


-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode {
  BOOL canMove;
  
  Particle *neighborParticle; 
  canMove = [super canMoveToNeighborCode:  neighborCode];

//Registriere Substratznzahl fr Rasmussen Regel 

 substratenumber  = [self countSubstrateNeighbors];  
 
 

  if (canMove) {
  neighborParticle = [self getAgentAtNeighborCode: neighborCode];
	
    if ([bondList getCount] > 0) {
    //Hier schauen wir, ob in die Richtung NeighborCode ein Bond existiert, Wenn ja, dann duerfen wir uns dahin nicht bewegen
      if ([self getBondAtNeighborCode: neighborCode] != nil)
	canMove = NO;
	
      //Hier wird kodiert, dass sich Links2 bevorzugt mit Substrate bewegen wenn diese existieren. (abgeschafft)!!!!
  //if([self countSubstrateNeighbors] > 0)
	//canMove = 
	  //([neighborParticle isMemberOf: [Substrate class]]);
	 //if([self countSubstrateNeighbors] == 0)
	 //canMove = 
	//([neighborParticle isMemberOf: [Hole class]] || 
 	 //  [neighborParticle isMemberOf: [Catalyst class]]);
  
   canMove = 
	([neighborParticle isMemberOf: [Hole class]] || 
	   [neighborParticle isMemberOf: [Link class]] || 
	   [neighborParticle isMemberOf: [Catalyst class]]|| 
	   [neighborParticle isMemberOf: [Substrate class]]);
  
  
  
    }
  else{
	//Hier wird kodiert, dass sich Links2 bevorzugt mit Substrate bewegen wenn diese existieren. (abgeschafft)!!!!
	//if([self countSubstrateNeighbors] > 0)
	//canMove = 
	 // ([neighborParticle isMemberOf: [Substrate class]]);
	// if([self countSubstrateNeighbors] == 0)
	 
	 canMove = 
	([neighborParticle isMemberOf: [Hole class]] || 
	   [neighborParticle isMemberOf: [Link class]] || 
	   [neighborParticle isMemberOf: [Catalyst class]]|| 
	   [neighborParticle isMemberOf: [Substrate class]]);
  
  }//Ende else       
      
  }
  
//printf(" [%d %d]",canMove, [bondList getCount] );
  return canMove;
}



-(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode{

//Link  laesst sich nur von Substrat bewegen ausser es gibt keines in seiner Nachbarschaft.


// Particle *neighborParticle;
BOOL canMove;


canMove =  [super canBeMovedToNeighborCode: neighborCode];

//diese Regeln sind jetzt nicht mehr ntig und wurden durch richtige Hydrophilie ersetzt
//neighborParticle = [self getAgentAtNeighborCode: neighborCode];
// if (![neighborParticle isMemberOf: [Substrate class]] && ([self countSubstrateNeighbors] > 0))
 //{ canMove = NO;}
  
//Wenn ein Link Bonds hat kann er nicht bewegt werden!

 if([self getBondCount] > 0) canMove = NO;

return canMove;


}



 
 



-(void) recordIfSplicable {
  NeighborRecord *neighborRecord;
  id bondableNeighbors;
  id index;

  bondableNeighbors = [List create: [self getZone]];
  [self scanForBondableNeighbors: bondableNeighbors];

  // Limit list to potential splices...
  index = [bondableNeighbors begin: [self getZone]];
  while ((neighborRecord = (NeighborRecord *) [index next]) != NULL) {
    if ([neighborRecord->particle getBondCount] != 1) {
      [index remove];
      [[self getZone] free: neighborRecord];
    }
  }
  [index drop];

  if ([bondList getCount] == 0)
   splicable = ([bondableNeighbors getCount] >= 2);
  else
   splicable = ([bondableNeighbors getCount] >= 1);

  [self dropNeighborList: bondableNeighbors];
}


-(void) countBondedNeighbors {
  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;

  bondedNeighborCount = 0U;
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Link2 class]]) {
	switch ([neighborParticle getBondCount]) {
	case 1:
	case 2:
	  bondedNeighborCount++;
	  break;
	default:
	}
      }
    } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
}





-(int) countSubstrateNeighbors {
  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice;
  int SubstrateCount = 0; 
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do {
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];

    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Substrate class]]) 
	  SubstrateCount++;
      } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
  
  return SubstrateCount;
}




 
-(BOOL) onlySubstrateNeighbors {
// Schaut ob ein Link2 nur von Substrat umringt ist oder nicht


  NeighborCode neighborCode, initialNeighborCode;
  id neighborParticle;
  id lattice; int i = 0;
  int SubstrateCount = 0; 
  lattice = [coord getLatticeObject];
  initialNeighborCode = [lattice getInitialNeighborCode];
  neighborCode = initialNeighborCode;
  do { i++;
    neighborParticle = [self getAgentAtNeighborCode: neighborCode];
      
    if (neighborParticle == nil) 
      [InternalError raiseEvent:
	 "nil encountered at neighborCode %d of particle %d.\n",
	 neighborCode, self];
    else {
      if ([neighborParticle isMemberOf: [Substrate class]]) 
	  SubstrateCount++;
      } 
    neighborCode = [lattice getNextNeighborCode: neighborCode];
  } while (neighborCode != initialNeighborCode);
  if(i>SubstrateCount)
  return NO;
  else return YES;
}



-(void) prepareForMotion {
  switch ([bondList getCount]) {

  case 0:
    if (link0SmartMotionFlag) {
      [self recordIfSplicable];
      [self countBondedNeighbors];
    }
    break;

  case 1:
    if (link1SmartMotionFlag) 
      [self recordIfSplicable];
    break;

  case 2:
  default:
  }
}


-(double) getMobilityFactor {
  double mobilityFactor;
  int bondCount;
  
//Wenn ein Link2 nur von Substrat umringt is, dann soll er normal sich bewegen.
  


  bondCount = [bondList getCount];

  switch (bondCount) {

  case 0: 
    mobilityFactor = link0MobilityFactor; 
    
    break;

  case 1: 
    mobilityFactor = link1MobilityFactor; 
       
    break;

  case 2: 
    mobilityFactor = link2MobilityFactor; 
    break;

  default:
      [InternalError raiseEvent:
       "Out-of-range bondCount (%d) on Link (%u).\n", 
       [bondList getCount], self];
      mobilityFactor = 0.0; // Defensive!
  }


  return mobilityFactor;
                                              




}


-(void) scanForRepairableNeighbors: (id) repairableList {
//   NeighborCode neighborCode, initialNeighborCode;
//   id neighborParticle;
//   id lattice;
// 
//   lattice = [coord getLatticeObject];
//   initialNeighborCode = [lattice getInitialNeighborCode];
//   neighborCode = initialNeighborCode;
//   do {
//     neighborParticle = [self getAgentAtNeighborCode: neighborCode];
// 
//     if (neighborParticle == nil) 
//       [InternalError raiseEvent:
// 	 "nil encountered at neighborCode %d of particle %d.\n",
// 	 neighborCode, self];
//     else {
//       if ([neighborParticle isMemberOf: [Link class]]) {
// 	if ([neighborParticle isDisintegrating] &&
// 	    ([neighborParticle getBondCount] > 0))
// 	  [repairableList addLast: neighborParticle];
//       }
//     } 
//     neighborCode = [lattice getNextNeighborCode: neighborCode];
//   } while (neighborCode != initialNeighborCode);
}

-(void) doRepairMotion {// 
//   id repairableLinkList;
//   id repairableLink;
// 
//   if ((!hasMoved) && (!disintegrating) && ([bondList getCount] == 0)) {
//     if ([bernoulli getTossWithProb: activeRepairProbability]) {
//       repairableLinkList = [List create: [self getZone]];
//       [self scanForRepairableNeighbors: repairableLinkList];
//       repairableLink = [self pickRandomListMember: repairableLinkList];
//       [repairableLinkList drop];
//       if (repairableLink != nil) {
// 	[repairableLink dropBonds];
// 	hasMoved = [self attemptSwapWithAgent: repairableLink];
// 	if (!hasMoved)
// 	  [InternalError raiseEvent:
// 	    "Attempted motion for repair failed?\n"];
// 	else {
// 	  [repairableLink setHasMoved];
// 	  [worldManager logRepairMotion];
// 	}
//       }
//     }
//   }
}


-(BOOL) displaceInDirection: (NeighborCode) displaceDirection {// 
//   id bond;
//   NeighborCode oldBondFromDirection, oldBondToDirection;
//   id oldBondPartner;
//   BOOL displaced;
// 
//   displaced = NO;
// 
//   if (disintegrating)
//     [InternalError raiseEvent:
//       "Attempted displacement, but disintegrating?\n"];
// 
//   if ([bondList getCount] != 2)
//     [InternalError raiseEvent:
//       "Attempted displacement, but (bondCount != 2)?\n"];
// 
//   if ([self canMoveToNeighborCode: displaceDirection]) {
//     if ([bernoulli getTossWithProb: 0.5]) 
//       bond = [bondList atOffset: 0];
//     else
//       bond = [bondList atOffset: 1];
// 
//     oldBondPartner = [bond getLinkOppositeTo: self];
//     oldBondFromDirection = [bond getNeighborCodeForLink: oldBondPartner];
//     oldBondToDirection = [bond getNeighborCodeForLink: self];
//     [bond clearLinks];
//     [bondMgr dropAgent: bond];
// 
//     hasMoved = [self attemptSwapToNeighborCode: displaceDirection];
//     if (hasMoved) {
//       [[self getAgentAtNeighborCode: displaceDirection] setHasMoved];
//       displaced = YES;
//     }
//     else {
//       // Reform the bond that was broken...
//       bond = [oldBondPartner createBondToNeighborCode: oldBondFromDirection];
//       [bond setLink1: self];
//       [bond setNeighbor1: oldBondToDirection];
//       [self addBond: bond];
//     }
//   }
//   return(displaced);
return(0);
}

-(void) scanForDisplaceableNeighbors: (id) displaceableNeighbors
  // displaceableNeighbors must be an empty List.
{
//   NeighborCode toNeighborCode, initialNeighborCode;
//   NeighborRecord *neighborRecord;
//   id neighborParticle;
//   id lattice;
// 
//   lattice = [coord getLatticeObject];
//   initialNeighborCode = [lattice getInitialNeighborCode];
//   toNeighborCode = initialNeighborCode;
//   do {
//     if ([self canMoveToNeighborCode: toNeighborCode]) {
//       neighborParticle = [self getAgentAtNeighborCode: toNeighborCode];
// 
//       if (neighborParticle == nil) 
// 	[InternalError raiseEvent:
// 	   "nil encountered at neighborCode %d of particle %d.\n",
// 	   toNeighborCode, self];
//       else {
// 	if ([neighborParticle isMemberOf: [Link class]]) {
// 	  if (([neighborParticle getBondCount] == 2) && 
//               (![neighborParticle isDisintegrating])) {
// 	    neighborRecord = [[self getZone] alloc: sizeof(NeighborRecord)];
// 	    neighborRecord->particle = neighborParticle;
// 	    neighborRecord->toCode = toNeighborCode;
// 	    neighborRecord->fromCode = 
//               [lattice getOppositeNeighborCode: toNeighborCode];
// 	    [displaceableNeighbors addLast: (id) neighborRecord];
// 	  }
// 	}
//       }
//     }
//     toNeighborCode = [lattice getNextNeighborCode: toNeighborCode];
//   } while (toNeighborCode != initialNeighborCode);
}


-(void) doDisplaceMotion {
 //  id displaceableLinkList;
//   NeighborRecord *neighborRecord;
//   NeighborCode displaceDirection;
//   id displaceableLink;
// 
//   if ((!hasMoved) && (!disintegrating) && 
//       ([bondList getCount] == 0) && (displaceProbability != 0.0)) {
//     if ([bernoulli getTossWithProb: displaceProbability]) {
//       displaceableLinkList = [List create: [self getZone]];
//       [self scanForDisplaceableNeighbors: displaceableLinkList];
//       neighborRecord = 
// 	(NeighborRecord *) [self pickRandomListMember: displaceableLinkList];
//       if (neighborRecord != NULL) {
// 	displaceableLink = neighborRecord->particle;
// 	displaceDirection = neighborRecord->toCode;
// 	if ([displaceableLink displaceInDirection: displaceDirection]) {
// 	  hasMoved = [self attemptSwapToNeighborCode: displaceDirection];
// 	  if (!hasMoved)
// 	    [InternalError raiseEvent:
// 	      "Attempted motion after displacement failed?\n"];
// 	  else {
// 	    [[self getAgentAtNeighborCode: displaceDirection] setHasMoved];  
// 	    [worldManager logDisplaceMotion];
// 	  }
// 	}
//       }
//       [self dropNeighborList: displaceableLinkList];
//     }
//   }
}

-attemptPromotion {
  id myCoord, cat;

  if (!absorbedSubstrate)
    {
      myCoord = [self copyCoord: [self getZone]];
      [link2Mgr dropAgent: self];
      cat = [catalystMgr createAgent];
      if (![cat attemptUnwarpToCoord: myCoord])
	[InternalError raiseEvent:
	  "Failed unwarp of cat %d.\n", link];
      [myCoord drop];
    }
  
  return self;
}

-doPromotion {
  if ((!promoting) &&
      (!disintegrating) && 
      ([bondList getCount] == 0) && 
      (linkPromotionProbability != 0.0))
    if ([self isNeighboringAgentOfClass: [Catalyst class]]) 
    // Promotion must be catalysed; the idea is to
    // prevent spontaneous promotion out in the
    // environment, which would just be a distraction and
    // possibly interfere with the cell proper...
      if ([bernoulli getTossWithProb: linkPromotionProbability]) {
        promoting = YES;
        forceEmission = absorbedSubstrate;
      }

  return self;
}

-step {
if([self getBondCount] > 1) [InternalError raiseEvent:
      "Link2 with more than 1 Bond"];

  //[self doRepairMotion];
  //[self doDisplaceMotion];
  [self doMotion];
  [self doBonding];
 //  [self doDisintegration];
//   [self doPromotion];
// 
//   if (forceEmission) {
//     [self attemptEmission];
//     forceEmission = absorbedSubstrate;
//   }
// 
//   if (disintegrating)
//     [self attemptDisintegration];
//   else 
//     if (promoting)
//       [self attemptPromotion];
//     else {
//       [self doEmission];
//       [self doAbsorption];
//    }

  return self;
}


-saveTo: (id <OutFile>) file {
  /* Note that we do *not* save bond information here; that info
     is already implicit in the bond agent list; of course, it
     would be more conservative to save the redundancies and then
     crosscheck on loading - but, for the time being at least,
     I judge that to be an unwarrananted complication... */

  [super saveTo: file];

  [file putInt: absorbedSubstrate];
  [file putString: " "];
  [file putInt: disintegrating];  
  [file putString: " "];

  return self;
}

-loadFrom: (id <InFile>) file toLattice: (id) inLattice {
  int intFlag;

  [super loadFrom: file toLattice: inLattice];

  [file getInt: &intFlag];
  if ((intFlag != YES) && (intFlag != NO))
    [InternalError raiseEvent:
      "Loaded BOOL value (%d) not YES or NO.", intFlag];
  absorbedSubstrate = (BOOL) intFlag; 


  [file getInt: &intFlag];
  if ((intFlag != YES) && (intFlag != NO))
    [InternalError raiseEvent:
      "Loaded BOOL value (%d) not YES or NO.", intFlag];
   disintegrating = (BOOL) intFlag;  

  return self;
}


@end
