/* 
  MooreNeighbor.m

  Barry McMullin <mcmullin@eeng.dcu.ie>
  JAN-1997

*/

#import <simtools.h>
#import "MooreNeighborhood.h"


@implementation MooreNeighborhood

-(void) setUniformUnsigned: (id) inUniformUnsigned {
  uniformUnsigned = inUniformUnsigned;
}

-(id) getUniformUnsigned {
  return uniformUnsigned;
}

-(OrientationCode) getOrientationCode: (NeighborCode) inNeighborCode {
  OrientationCode orientationCode;

  switch (inNeighborCode) {
  case SqrNorth:     
  case SqrSouth:     orientationCode = SqrNorthSouth;         break;
  case SqrEast:     
  case SqrWest:      orientationCode = SqrEastWest;           break;
  case SqrNorthEast:     
  case SqrSouthWest: orientationCode = SqrNorthEastSouthWest; break;
  case SqrNorthWest:     
  case SqrSouthEast: orientationCode = SqrNorthWestSouthEast; break;
  default:
    [InternalError raiseEvent:
     "Invalid Moore neighbor code (%d) encountered.\n", 
     inNeighborCode];
    orientationCode = SqrNorthSouth; // Defensive.
  }

  return orientationCode;
}


-(NeighborCode) getInitialNeighborCode {
  return SqrNorth;
}

-(char *) getNeighborCodeName: (NeighborCode) inNeighborCode {
  char *name;

  switch (inNeighborCode) {
  case SqrNorth:     name = "SqrNorth";     break;
  case SqrNorthEast: name = "SqrNorthEast"; break;
  case SqrEast:      name = "SqrEast";      break;
  case SqrSouthEast: name = "SqrSouthEast"; break;
  case SqrSouth:     name = "SqrSouth";     break;
  case SqrSouthWest: name = "SqrSouthWest"; break;
  case SqrWest:      name = "SqrWest";      break;
  case SqrNorthWest: name = "SqrNorthWest"; break;
  default:
    [InternalError raiseEvent:
     "Invalid Moore neighbor code (%d) encountered.\n", 
     inNeighborCode];
    name = "Invalid"; // Defensive...
  }

  return (name);
}

-(BOOL) isValidNeighborCode: (NeighborCode) inNeighborCode {
  BOOL isValid;

  switch (inNeighborCode) {
    case SqrNorth: 
    case SqrNorthEast:
    case SqrEast:  
    case SqrSouthEast:
    case SqrSouth: 
    case SqrSouthWest:
    case SqrWest:  
    case SqrNorthWest:
      isValid = YES;
      break;
    default:
      isValid = NO;
  }

  return(isValid);
}


-(NeighborCode) getNextNeighborCode: (NeighborCode) inNeighborCode {
  NeighborCode nextNeighborCode;

  switch (inNeighborCode) {
  case SqrNorth:     nextNeighborCode = SqrNorthEast;    break;
  case SqrNorthEast: nextNeighborCode = SqrEast;         break;
  case SqrEast:      nextNeighborCode = SqrSouthEast;    break;
  case SqrSouthEast: nextNeighborCode = SqrSouth;        break;
  case SqrSouth:     nextNeighborCode = SqrSouthWest;    break;
  case SqrSouthWest: nextNeighborCode = SqrWest;            break;
  case SqrWest:      nextNeighborCode = SqrNorthWest;    break;
  case SqrNorthWest: nextNeighborCode = SqrNorth;        break;
  default:
    [InternalError raiseEvent:
     "Invalid Moore neighbor code (%d) encountered.\n", 
     inNeighborCode];
    nextNeighborCode = SqrNorth; // Defensive...
  }

  return (nextNeighborCode);
}

-(NeighborCode) getPreviousNeighborCode: (NeighborCode) inNeighborCode {
  NeighborCode previousNeighborCode;

  switch (inNeighborCode) {
  case SqrNorth:     previousNeighborCode = SqrNorthWest;    break;
  case SqrNorthWest: previousNeighborCode = SqrWest;         break;
  case SqrWest:      previousNeighborCode = SqrSouthWest;    break;
  case SqrSouthWest: previousNeighborCode = SqrSouth;        break;
  case SqrSouth:     previousNeighborCode = SqrSouthEast;    break;
  case SqrSouthEast: previousNeighborCode = SqrEast;         break;
  case SqrEast:      previousNeighborCode = SqrNorthEast;    break;
  case SqrNorthEast: previousNeighborCode = SqrNorth;        break;
  default:
    [InternalError raiseEvent:
     "Invalid Moore neighbor code (%d) encountered.\n", 
      inNeighborCode];
    previousNeighborCode = SqrNorth; // Defensive...
  }

  return (previousNeighborCode);
}

-(NeighborCode) getOppositeNeighborCode: (NeighborCode) inNeighborCode {
  NeighborCode oppositeNeighborCode;

  switch (inNeighborCode) {
  case SqrNorth:     oppositeNeighborCode = SqrSouth;        break;
  case SqrNorthEast: oppositeNeighborCode = SqrSouthWest;    break;
  case SqrWest:      oppositeNeighborCode = SqrEast;         break;
  case SqrNorthWest: oppositeNeighborCode = SqrSouthEast;    break;
  case SqrSouth:     oppositeNeighborCode = SqrNorth;        break;
  case SqrSouthEast: oppositeNeighborCode = SqrNorthWest;    break;
  case SqrEast:      oppositeNeighborCode = SqrWest;         break;
  case SqrSouthWest: oppositeNeighborCode = SqrNorthEast;    break;
  default:
    [InternalError raiseEvent:
     "Invalid Moore neighbor code (%d) encountered.\n", 
      inNeighborCode];
    oppositeNeighborCode = SqrNorth; // Defensive...
  }

  return (oppositeNeighborCode);

}

-(NeighborCode) getRandomNeighborCode {
  unsigned randomNumber;
  NeighborCode randomNeighborCode;

  randomNeighborCode = SqrNorth; // Defensive...

  if (uniformUnsigned == nil)
    [InternalError raiseEvent:
      "Attempt to -getRandomNeighborCode, but uniformUnsigned is nil.\n"];
  else {
    randomNumber = [uniformUnsigned getUnsignedWithMin: 0 withMax: 7];
    switch (randomNumber) {
    case 0: randomNeighborCode = SqrNorth;        break;
    case 1: randomNeighborCode = SqrNorthEast;    break;
    case 2: randomNeighborCode = SqrEast;         break;
    case 3: randomNeighborCode = SqrSouthEast;    break;
    case 4: randomNeighborCode = SqrSouth;        break;
    case 5: randomNeighborCode = SqrSouthWest;    break;
    case 6: randomNeighborCode = SqrWest;         break;
    case 7: randomNeighborCode = SqrNorthWest;    break;
    default:
      [InternalError raiseEvent:
       "Out-of-range random number (%d) encountered.\n", randomNumber];
    }
  }

  return (randomNeighborCode);
}

-(NeighborCode) getRandomNeighborCodeExcluding: (NeighborCode) inNeighborCode {
  // This is a cheap `n cheerful, totally unoptimised version...
  unsigned randomNumber;
  NeighborCode randomNeighborCode;

  randomNeighborCode = [self getNextNeighborCode: inNeighborCode];

  if (uniformUnsigned == nil)
    [InternalError raiseEvent:
      "Attempt to -getRandomNeighborCodeExcluding,\n"
       "but uniformUnsigned is nil.\n"];
  else {
    randomNumber = [uniformUnsigned getUnsignedWithMin: 0 withMax: 6];
    while (randomNumber > 0) {
      randomNeighborCode = [self getNextNeighborCode: randomNeighborCode];
      randomNumber--;
    }
  }

  return (randomNeighborCode);
}


@end
