/* 
  Neighborhood.h

  Barry McMullin <mcmullin@eeng.dcu.ie>
  JAN-1997

*/

#import <objectbase.h>
#import <random.h>

#import "NeighborCode.h"
#import "OrientationCode.h"

@protocol Neighborhood

/* The precise meanings of -next and -previous are specific
   to a particular neighborhood class.

  The -random method picks a random neighbor (with 
  uniform probability function) and returns it.  */

-(BOOL) isValidNeighborCode: (NeighborCode) inNeighbor;
-(OrientationCode) getOrientationCode: (NeighborCode) inNeighborCode;
-(NeighborCode) getInitialNeighborCode;
-(NeighborCode) getNextNeighborCode: (NeighborCode) inNeighbor;
-(NeighborCode) getPreviousNeighborCode: (NeighborCode) inNeighbor;
-(NeighborCode) getOppositeNeighborCode: (NeighborCode) inNeighbor;

-(char *) getNeighborCodeName: (NeighborCode) inNeighbor;

-(void) setUniformUnsigned: (id) inUniformUnsigned;
-(id) getUniformUnsigned;
/* If you wish to use any of the randomizing methods of
  the protocol, a uniformUnsigned distribution object
  (of the "variable" or non-"frozen" kind) must be
  available to the Neighborhood object.  This must be
  provided from outside via -setUniformUnsigned. This
  approach is preferred to having the Neighborhood
  object create its own private distribution object,
  because it will be quite common to create a
  uniformUnsigned distribution object for more general
  use in an app anyway, and share it around whoever
  else wants to use it...
*/

-(NeighborCode) getRandomNeighborCode;
-(NeighborCode) getRandomNeighborCodeExcluding: (NeighborCode) inNeighbor;

@end
