/* 
  Particle.h

  Barry McMullin <mcmullin@eeng.dcu.ie>
  27-SEP-1996

  This is a general purpose class for "particles" in a 2d
  space. It should be subclassed for distinct "molecular
  species" in an artificial chemistry.

*/

#import <tkobjc/Raster.h>

#import "NeighborCode.h"
#import "LatticeAgent.h"


@interface Particle: LatticeAgent {
  BOOL hasMoved;
}

+(void) setClassMobilityFactor: (double) inMobilityFactor;
+(double) getClassMobilityFactor;


/* This Bernoulli distribution is used in implementing
   the random walk style of motion exhibited (by
   default) by particles.  It must be a "variable" or
   non-"frozen" flavour of Bernoulli, created
   externally to the particles; in this way a single
   bernoulli object can be shared around multiple
   usages in a single app.  */
+(void) setBernoulli: (id) inBernoulli; 
+(id) getBernoulli;

/* This UniformUnsigned distribution is used in implementing
   the *energetic* style motion only. It must be a "variable" or
   non-"frozen" flavour of UnsignedUniform, created
   externally to the particles; in this way a single
   unsignedUniform object can be shared around multiple
   usages in a single app.  */
+(void) setUniformUnsigned: (id) inUniformUnsigned; 
+(id) getUniformUnsigned;

+(void) setEnergeticMotionFlag: (BOOL) value; 
+(BOOL) getEnergeticMotionFlag;

-(double) getMobilityFactor;
-(BOOL) canMoveToNeighborCode: (NeighborCode) neighborCode;
-(BOOL) canBeMovedToNeighborCode: (NeighborCode) neighborCode;
-doMotion;
-setHasMoved;
-initialiseStep;
-step;


@end
